#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import torch
from typing import Callable, Dict, List, Optional, Sequence, Tuple
from pathlib import Path
from pytorch_lightning import LightningModule
from torch import Tensor, argmax, mode, nn, optim, round
from pytorch_lightning.utilities.rank_zero import rank_zero_warn
from torchmetrics.classification import (
    MulticlassAUROC,
    MulticlassAccuracy,
    MulticlassConfusionMatrix,
    MulticlassCohenKappa,
    MulticlassAveragePrecision,
    BinaryConfusionMatrix,
    BinaryAccuracy,
    BinaryPrecision,
    BinaryRecall,
    BinaryF1Score,
    BinaryCohenKappa,
    BinaryAUROC,
    BinarySpecificity,
    BinaryAveragePrecision,
)
from health_ml.utils import log_on_epoch
from health_ml.deep_learning_config import OptimizerParams
from health_cpath.models.encoders import IdentityEncoder
from health_cpath.utils.deepmil_utils import ClassifierParams, EncoderParams, PoolingParams
from health_cpath.datasets.base_dataset import TilesDataset
from health_cpath.utils.naming import DeepMILSubmodules, MetricsKey, ResultsKey, SlideKey, ModelKey, TileKey
from health_cpath.utils.output_utils import (
    BatchResultsType,
    DeepMILOutputsHandler,
    EpochResultsType,
    validate_class_names,
    EXTRA_PREFIX,
)


RESULTS_COLS = [
    ResultsKey.SLIDE_ID,
    ResultsKey.TILE_ID,
    ResultsKey.IMAGE_PATH,
    ResultsKey.PROB,
    ResultsKey.CLASS_PROBS,
    ResultsKey.PRED_LABEL,
    ResultsKey.TRUE_LABEL,
    ResultsKey.BAG_ATTN,
]


def _format_cuda_memory_stats() -> str:
    return (
        f"GPU {torch.cuda.current_device()} memory: "
        f"{torch.cuda.memory_allocated() / 1024 ** 3:.2f} GB allocated, "
        f"{torch.cuda.memory_reserved() / 1024 ** 3:.2f} GB reserved"
    )


class DeepMILModule(LightningModule):
    """Base class for deep multiple-instance learning"""

    def __init__(
        self,
        label_column: str,
        n_classes: int,
        class_weights: Optional[Tensor] = None,
        class_names: Optional[Sequence[str]] = None,
        encoder_params: EncoderParams = EncoderParams(),
        pooling_params: PoolingParams = PoolingParams(),
        classifier_params: ClassifierParams = ClassifierParams(),
        optimizer_params: OptimizerParams = OptimizerParams(),
        outputs_folder: Optional[Path] = None,
        outputs_handler: Optional[DeepMILOutputsHandler] = None,
        analyse_loss: Optional[bool] = False,
        verbose: bool = False,
    ) -> None:
        """
        :param label_column: Label key for input batch dictionary.
        :param n_classes: Number of output classes for MIL prediction. For binary classification, n_classes should be
         set to 1.
        :param class_weights: Tensor containing class weights (default=None).
        :param class_names: The names of the classes if available (default=None).
        :param encoder_params: Encoder parameters that specify all encoder specific attributes.
        :param pooling_params: Pooling layer parameters that specify all encoder specific attributes.
        :param classifier_params: Classifier parameters that specify all classifier specific attributes.
        :param optimizer_params: Optimizer parameters that specify all specific attributes to be used for oprimization.
        :param outputs_folder: Path to output folder where encoder checkpoint is downloaded.
        :param outputs_handler: A configured :py:class:`DeepMILOutputsHandler` object to save outputs for the best
            validation epoch and test stage. If omitted (default), no outputs will be saved to disk (aside from usual
            metrics logging).
        :param analyse_loss: If True, the loss is analysed per sample and analysed with LossAnalysisCallback.
        :param verbose: if True statements about memory usage are output at each step.
        """
        super().__init__()

        # Dataset specific attributes
        self.label_column = label_column
        self.n_classes = n_classes
        self.class_weights = class_weights
        self.class_names = validate_class_names(class_names, self.n_classes)

        self.encoder_params = encoder_params
        self.pooling_params = pooling_params
        self.classifier_params = classifier_params
        self.optimizer_params = optimizer_params

        self.save_hyperparameters()
        self.verbose = verbose
        self.outputs_handler = outputs_handler

        # This flag can be switched on before invoking trainer.validate() to enable saving additional time/memory
        # consuming validation outputs via calling self.on_run_extra_validation_epoch()
        self._on_extra_val_epoch = False

        # Model components
        self.encoder = encoder_params.get_encoder(outputs_folder)
        self.projector = encoder_params.get_projection_layer(self.encoder.num_encoding)
        # If projection is enabled, the number of encoding dimensions is the projection dimension otherwise it is the
        # number of encoding dimensions of the encoder.
        num_encoding = encoder_params.projection_dim if encoder_params.projection_dim > 0 else self.encoder.num_encoding
        self.aggregation_fn, self.num_pooling = pooling_params.get_pooling_layer(num_encoding)
        self.classifier_fn = classifier_params.get_classifier(self.num_pooling, self.n_classes)
        self.activation_fn = self.get_activation()

        # Loss function
        self.loss_fn = self.get_loss(reduction="mean")
        self.loss_fn_no_reduction = self.get_loss(reduction="none")
        self.analyse_loss = analyse_loss

        # Metrics Objects
        self.train_metrics = self.get_metrics()
        self.val_metrics = self.get_metrics()
        self.test_metrics = self.get_metrics()

        self.reset_encoder_to_identity_if_caching()

    def reset_encoder_to_identity_if_caching(self) -> None:
        """If caching is enabled, the encoder is replaced with an identity encoder."""
        if self.encoder_params.is_caching:
            # Encoding is done in the datamodule, so here we provide instead a dummy
            # no-op IdentityEncoder to be used inside the model
            self.encoder = IdentityEncoder(input_dim=(self.encoder.num_encoding,))

    @staticmethod
    def copy_weights(
        current_submodule: nn.Module, pretrained_submodule: nn.Module, submodule_name: DeepMILSubmodules
    ) -> None:
        """Copy weights from pretrained submodule to current submodule.

        :param current_submodule: Submodule to copy weights to.
        :param pretrained_submodule: Submodule to copy weights from.
        :param submodule_name: Name of the submodule.
        """

        def _total_params(submodule: nn.Module) -> int:
            return sum(p.numel() for p in submodule.parameters())

        pre_total_params = _total_params(pretrained_submodule)
        cur_total_params = _total_params(current_submodule)

        if pre_total_params != cur_total_params:
            raise ValueError(
                f"Submodule {submodule_name} has different number of parameters "
                f"({cur_total_params} vs {pre_total_params}) from pretrained model."
            )

        for param, pretrained_param in zip(
            current_submodule.state_dict().values(), pretrained_submodule.state_dict().values()
        ):
            try:
                param.data.copy_(pretrained_param.data)
            except Exception as e:
                raise ValueError(f"Failed to copy weights for {submodule_name} because of the following exception: {e}")

    def transfer_weights(self, pretrained_checkpoint_path: Optional[Path]) -> None:
        """Transfer weights from pretrained checkpoint if provided."""

        if pretrained_checkpoint_path:
            pretrained_model = self.load_from_checkpoint(checkpoint_path=str(pretrained_checkpoint_path))

            if self.encoder_params.pretrained_encoder:
                self.copy_weights(self.encoder, pretrained_model.encoder, DeepMILSubmodules.ENCODER)

            if self.pooling_params.pretrained_pooling:
                self.copy_weights(self.aggregation_fn, pretrained_model.aggregation_fn, DeepMILSubmodules.POOLING)

            if self.classifier_params.pretrained_classifier:
                if pretrained_model.n_classes != self.n_classes:
                    raise ValueError(
                        f"Number of classes in pretrained model {pretrained_model.n_classes} "
                        f"does not match number of classes in current model {self.n_classes}."
                    )
                self.copy_weights(self.classifier_fn, pretrained_model.classifier_fn, DeepMILSubmodules.CLASSIFIER)

    def get_loss(self, reduction: str = "mean") -> Callable:
        if self.n_classes > 1:
            if self.class_weights is None:
                return nn.CrossEntropyLoss(reduction=reduction)
            else:
                class_weights = self.class_weights.float()
                return nn.CrossEntropyLoss(weight=class_weights, reduction=reduction)
        else:
            pos_weight = None
            if self.class_weights is not None:
                pos_weight = Tensor([self.class_weights[1] / (self.class_weights[0] + 1e-5)])
            return nn.BCEWithLogitsLoss(pos_weight=pos_weight, reduction=reduction)

    def get_activation(self) -> Callable:
        if self.n_classes > 1:
            return nn.Softmax(dim=-1)
        else:
            return nn.Sigmoid()

    @staticmethod
    def get_bag_label(labels: Tensor) -> Tensor:
        """Get bag label as the majority class of the bag's samples. For slides pipeline, we already have a single
        label per bag so we just return that label. For tiles pipeline, we need to get the majority class as labels
        are duplicated for each tile in the bag."""
        if len(labels.shape) == 0:
            return labels
        bag_label = mode(labels).values
        return bag_label.view(1)

    def get_metrics(self) -> nn.ModuleDict:
        if self.n_classes > 1:
            return nn.ModuleDict(
                {
                    MetricsKey.ACC: MulticlassAccuracy(num_classes=self.n_classes, average='micro'),
                    MetricsKey.AUROC: MulticlassAUROC(num_classes=self.n_classes),
                    MetricsKey.AVERAGE_PRECISION: MulticlassAveragePrecision(num_classes=self.n_classes),
                    # Quadratic Weighted Kappa (QWK) used in PANDA challenge
                    # is calculated using Cohen's Kappa with quadratic weights
                    # https://www.kaggle.com/code/reighns/understanding-the-quadratic-weighted-kappa/
                    MetricsKey.COHENKAPPA: MulticlassCohenKappa(num_classes=self.n_classes, weights='quadratic'),
                    MetricsKey.CONF_MATRIX: MulticlassConfusionMatrix(num_classes=self.n_classes),
                    # Metrics below are computed for multi-class case only
                    MetricsKey.ACC_MACRO: MulticlassAccuracy(num_classes=self.n_classes, average='macro'),
                    MetricsKey.ACC_WEIGHTED: MulticlassAccuracy(num_classes=self.n_classes, average='weighted'),
                }
            )
        else:
            return nn.ModuleDict(
                {
                    MetricsKey.ACC: BinaryAccuracy(),
                    MetricsKey.AUROC: BinaryAUROC(),
                    # Average precision is a measure of area under the PR curve
                    MetricsKey.AVERAGE_PRECISION: BinaryAveragePrecision(),
                    MetricsKey.COHENKAPPA: BinaryCohenKappa(weights='quadratic'),
                    MetricsKey.CONF_MATRIX: BinaryConfusionMatrix(),
                    # Metrics below are computed for binary case only
                    MetricsKey.F1: BinaryF1Score(),
                    MetricsKey.PRECISION: BinaryPrecision(),
                    MetricsKey.RECALL: BinaryRecall(),
                    MetricsKey.SPECIFICITY: BinarySpecificity(),
                }
            )

    def get_extra_prefix(self) -> str:
        """Get extra prefix for the metrics name to avoir overriding best validation metrics."""
        return EXTRA_PREFIX if self._on_extra_val_epoch else ""

    def log_metrics(self, stage: str, prefix: str = '') -> None:
        valid_stages = set([stage for stage in ModelKey])
        if stage not in valid_stages:
            raise Exception(f"Invalid stage. Chose one of {valid_stages}")
        for metric_name, metric_object in self.get_metrics_dict(stage).items():
            if metric_name == MetricsKey.CONF_MATRIX:
                metric_value = metric_object.compute()
                metric_value_n = metric_value / metric_value.sum(axis=1, keepdims=True)
                for i in range(metric_value_n.shape[0]):
                    log_on_epoch(self, f'{prefix}{stage}/{self.class_names[i]}', metric_value_n[i, i])
            else:
                log_on_epoch(self, f'{prefix}{stage}/{metric_name}', metric_object)

    def get_instance_features(self, instances: Tensor) -> Tensor:
        if not self.encoder_params.tune_encoder:
            self.encoder.eval()
        if self.encoder_params.encoding_chunk_size > 0:
            embeddings = []
            chunks = torch.split(instances, self.encoder_params.encoding_chunk_size)
            for chunk in chunks:
                chunk_embeddings = self.encoder(chunk)
                embeddings.append(chunk_embeddings)
            instance_features = torch.cat(embeddings)
        else:
            instance_features = self.encoder(instances)  # N X L x 1 x 1
        return instance_features

    def get_attentions_and_bag_features(self, instance_features: Tensor) -> Tuple[Tensor, Tensor]:
        if not self.pooling_params.tune_pooling:
            self.aggregation_fn.eval()
        attentions, bag_features = self.aggregation_fn(instance_features)  # K x N | K x L
        bag_features = bag_features.view(1, -1)
        return attentions, bag_features

    def get_bag_logit(self, bag_features: Tensor) -> Tensor:
        if not self.classifier_params.tune_classifier:
            self.classifier_fn.eval()
        bag_logit = self.classifier_fn(bag_features)
        return bag_logit

    def forward(self, instances: Tensor) -> Tuple[Tensor, Tensor]:  # type: ignore
        instance_features = self.get_instance_features(instances)
        projected_features = self.projector(instance_features)
        attentions, bag_features = self.get_attentions_and_bag_features(projected_features)
        bag_logit = self.get_bag_logit(bag_features)
        return bag_logit, attentions

    def configure_optimizers(self) -> optim.Optimizer:
        return optim.Adam(
            filter(lambda p: p.requires_grad, self.parameters()),
            lr=self.optimizer_params.l_rate,
            weight_decay=self.optimizer_params.weight_decay,
            betas=self.optimizer_params.adam_betas,
        )

    def get_metrics_dict(self, stage: str) -> nn.ModuleDict:
        return getattr(self, f'{stage}_metrics')

    def compute_bag_labels_logits_and_attn_maps(self, batch: Dict) -> Tuple[Tensor, Tensor, List]:
        # The batch dict contains lists of tensors of different sizes, for all bags in the batch.
        # This means we can't stack them along a new axis without padding to the same length.
        # We could alternatively concatenate them, but this would require other changes (e.g. in
        # the attention layers) to correctly split the tensors by bag/slide ID.
        bag_labels_list = []
        bag_logits_list = []
        bag_attn_list = []
        for bag_idx in range(len(batch[self.label_column])):
            images = batch[TilesDataset.IMAGE_COLUMN][bag_idx].float()
            labels = batch[self.label_column][bag_idx]
            bag_labels_list.append(self.get_bag_label(labels))
            logit, attn = self(images)
            bag_logits_list.append(logit.view(-1))
            bag_attn_list.append(attn)
        bag_logits = torch.stack(bag_logits_list)
        bag_labels = torch.stack(bag_labels_list).view(-1)
        return bag_logits, bag_labels, bag_attn_list

    def update_results_with_metadata(self, batch: Dict, results: Dict) -> None:
        """Update results with metadata. This can be either tiles or slides metadata including tiles coordinates."""
        results.update({ResultsKey.SLIDE_ID: batch[SlideKey.SLIDE_ID], ResultsKey.TILE_ID: batch[TileKey.TILE_ID]})
        # Add tiles coordinates if available
        coordinates_keys = [TileKey.TILE_TOP, TileKey.TILE_BOTTOM, TileKey.TILE_RIGHT, TileKey.TILE_LEFT]
        if all([key in batch for key in coordinates_keys]):
            for key in coordinates_keys:
                results[key] = batch[key]
        elif TilesDataset.TILE_X_COLUMN in batch and TilesDataset.TILE_X_COLUMN in batch:
            results[ResultsKey.TILE_LEFT] = batch[TilesDataset.TILE_X_COLUMN]
            results[ResultsKey.TILE_TOP] = batch[TilesDataset.TILE_Y_COLUMN]
        else:
            rank_zero_warn(
                message="Coordinates not found in batch. If this is not expected check your input tiles dataset."
            )

    def update_slides_selection(self, stage: str, batch: Dict, results: Dict) -> None:
        if (
            (stage == ModelKey.TEST or (stage == ModelKey.VAL and self._on_extra_val_epoch))
            and self.outputs_handler
            and self.outputs_handler.tiles_selector
        ):
            self.outputs_handler.tiles_selector.update_slides_selection(batch, results)

    def _compute_loss(self, loss_fn: Callable, bag_logits: Tensor, bag_labels: Tensor) -> Tensor:
        if self.n_classes > 1:
            return loss_fn(bag_logits, bag_labels.long())
        else:
            return loss_fn(bag_logits.squeeze(1), bag_labels.float())

    def _shared_step(self, batch: Dict, batch_idx: int, stage: str) -> BatchResultsType:
        bag_logits, bag_labels, bag_attn_list = self.compute_bag_labels_logits_and_attn_maps(batch)

        loss = self._compute_loss(self.loss_fn, bag_logits, bag_labels)

        predicted_probs = self.activation_fn(bag_logits)
        if self.n_classes > 1:
            predicted_labels = argmax(predicted_probs, dim=1)
            probs_perclass = predicted_probs
        else:
            predicted_labels = round(predicted_probs).int()
            probs_perclass = Tensor(
                [
                    [1.0 - predicted_probs[i][0].item(), predicted_probs[i][0].item()]
                    for i in range(len(predicted_probs))
                ]
            )

        loss = loss.view(-1, 1)
        predicted_labels = predicted_labels.view(-1, 1)
        batch_size = predicted_labels.shape[0]

        if self.n_classes == 1:
            predicted_probs = predicted_probs.squeeze(dim=1)

        results = dict()

        if self.analyse_loss and stage in [ModelKey.TRAIN, ModelKey.VAL]:
            loss_per_sample = self._compute_loss(self.loss_fn_no_reduction, bag_logits, bag_labels)
            results[ResultsKey.LOSS_PER_SAMPLE] = loss_per_sample.detach().cpu().numpy()

        bag_labels = bag_labels.view(-1, 1)

        for metric_object in self.get_metrics_dict(stage).values():
            metric_object.update(
                predicted_probs,
                bag_labels.view(
                    batch_size,
                ).int(),
            )
        results.update(
            {
                ResultsKey.LOSS: loss,
                ResultsKey.PROB: predicted_probs,
                ResultsKey.CLASS_PROBS: probs_perclass,
                ResultsKey.PRED_LABEL: predicted_labels,
                ResultsKey.TRUE_LABEL: bag_labels.int(),
                ResultsKey.BAG_ATTN: bag_attn_list,
            }
        )
        self.update_results_with_metadata(batch=batch, results=results)
        self.update_slides_selection(stage=stage, batch=batch, results=results)
        return results

    def training_step(self, batch: Dict, batch_idx: int) -> BatchResultsType:  # type: ignore
        train_result = self._shared_step(batch, batch_idx, ModelKey.TRAIN)
        self.log('train/loss', train_result[ResultsKey.LOSS], on_epoch=True, on_step=True, logger=True, sync_dist=True)
        if self.verbose:
            print(f"After loading images batch {batch_idx} -", _format_cuda_memory_stats())
        results = {ResultsKey.LOSS: train_result[ResultsKey.LOSS]}
        if self.analyse_loss:
            results.update(
                {
                    ResultsKey.LOSS_PER_SAMPLE: train_result[ResultsKey.LOSS_PER_SAMPLE],
                    ResultsKey.CLASS_PROBS: train_result[ResultsKey.CLASS_PROBS],
                    ResultsKey.TILE_ID: train_result[ResultsKey.TILE_ID],
                }
            )
        return results

    def validation_step(self, batch: Dict, batch_idx: int) -> BatchResultsType:  # type: ignore
        val_result = self._shared_step(batch, batch_idx, ModelKey.VAL)
        name = f'{self.get_extra_prefix()}val/loss'
        self.log(name, val_result[ResultsKey.LOSS], on_epoch=True, on_step=True, logger=True, sync_dist=True)
        return val_result

    def test_step(self, batch: Dict, batch_idx: int) -> BatchResultsType:  # type: ignore
        test_result = self._shared_step(batch, batch_idx, ModelKey.TEST)
        self.log('test/loss', test_result[ResultsKey.LOSS], on_epoch=True, on_step=True, logger=True, sync_dist=True)
        return test_result

    def training_epoch_end(self, outputs: EpochResultsType) -> None:  # type: ignore
        self.log_metrics(ModelKey.TRAIN)

    def validation_epoch_end(self, epoch_results: EpochResultsType) -> None:  # type: ignore
        self.log_metrics(stage=ModelKey.VAL, prefix=self.get_extra_prefix())
        if self.outputs_handler:
            self.outputs_handler.save_validation_outputs(
                epoch_results=epoch_results,
                metrics_dict=self.get_metrics_dict(ModelKey.VAL),  # type: ignore
                epoch=self.current_epoch,
                is_global_rank_zero=self.global_rank == 0,
                on_extra_val=self._on_extra_val_epoch,
            )

    def test_epoch_end(self, epoch_results: EpochResultsType) -> None:  # type: ignore
        self.log_metrics(ModelKey.TEST)
        if self.outputs_handler:
            self.outputs_handler.save_test_outputs(
                epoch_results=epoch_results, is_global_rank_zero=self.global_rank == 0
            )

    def on_run_extra_validation_epoch(self) -> None:
        """Hook to be called at the beginning of an extra validation epoch to set validation plots options to the same
        as the test plots options."""
        self._on_extra_val_epoch = True
        if self.outputs_handler:
            self.outputs_handler.val_plots_handler.plot_options = self.outputs_handler.test_plots_handler.plot_options



#  -------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  -------------------------------------------------------------------------------------------
from enum import Enum
import logging
import math
import numpy as np

from health_cpath.preprocessing.loading import WSIBackend
from health_cpath.utils.naming import SlideKey
from monai.data.wsi_reader import WSIReader
from monai.transforms import MapTransform
from openslide import OpenSlide
from pathlib import Path
from tifffile.tifffile import TiffWriter, PHOTOMETRIC, COMPRESSION
from typing import Any, Dict, List, Optional, Tuple


logger = logging.getLogger(__name__)

AMPERSAND = "&"
UNDERSCORE = "_"
TIFF_EXTENSION = ".tiff"


class ResolutionUnit(str, Enum):
    """The unit of the resolution of the tiff file. This is used to calculate the resolution of the tiff file."""

    INCH = "inch"
    CENTIMETER = "centimeter"
    MILLIMETER = "millimeter"
    MICROMETER = "micrometer"


class ConvertWSIToTiffd(MapTransform):
    """Converts a wsi file to a tiff file. The tiff file is saved in the output_folder with the same name as the src
    file but with the tiff extension. Ampersands are replaced by the replace_ampersand_by string. The tiff file
    contains the image data at the target magnifications. If target_magnifications is None, the tiff file contains the
    image data at all magnifications. If add_lowest_magnification is True, the tiff file also contains the image data
    at the lowest magnification. The tiff file is saved with the compression specified by the compression parameter with
    a fixed tile size. This works with all supported wsi formats by openslide.
    """

    OBJECTIVE_POWER_KEY = "openslide.objective-power"
    RESOLUTION_UNIT_KEY = "tiff.ResolutionUnit"
    RESOLUTION_UNIT = ResolutionUnit.CENTIMETER
    SOFTWARE = "tifffile"

    def __init__(
        self,
        output_folder: Path,
        image_key: str = SlideKey.IMAGE,
        target_magnifications: Optional[List[float]] = [10.0],
        add_lowest_magnification: bool = False,
        default_base_objective_power: Optional[float] = None,
        replace_ampersand_by: str = UNDERSCORE,
        compression: COMPRESSION = COMPRESSION.ADOBE_DEFLATE,
        tile_size: int = 512,
        min_file_size: int = 0,
        verbose: bool = False,
    ) -> None:
        """
        :param output_folder: The directory where the tiff file will be saved.
        :param image_key: The key of the image that should be converted in the data dictionary, defaults to
            `SlideKey.IMAGE`.
        :param target_magnifications: The magnifications that should be read, converted and written to the output file,
            e.g. [10., 20.], defaults to [10.]. If target_magnifications is None, the tiff file will contain the image
            data at all magnifications.
        :param add_lowest_magnification: A flag indicating whether the tiff file should also contain the image data at
            the lowest magnification, defaults to False. This is useful if the lowest magnification of the wsi is not
            part of the target magnifications and one wants to use the lowest magnification for faster processing.
        :param default_base_objective_power: The base objective power of the wsi. This is used to calculate the
            magnification of the wsi. If the objective power is not found in the wsi properties, the
            base_objective_power is used instead. If the objective power is not found in the wsi properties and
            base_objective_power is None, an error is raised, defaults to None.
        :param replace_ampersand_by: A string that is used to replace ampersands in the src file name, defaults to
            `UNDERSCORE`. This is useful because ampersands in file names can cause problems in cloud storage.
        :param compression: The compression that is used to save the tiff file, defaults to `COMPRESSION.ADOBE_DEFLATE`
            aka ZLIB that is lossless compression. Make sure to use one of these options (RAW, LZW, JPEG, JPEG2000) so
            that the converted files are readable by cucim.
        :param tile_size: The size of the tiles that are used to write the tiff file, defaults to 512.
        :param min_file_size: The minimum size of the tiff file in bytes. If the tiff file is smaller than this size, it
            will get overwritten. Defaults to 0.
        :param verbose: A flag to enable verbose logging, defaults to False.
        """
        self.output_folder = output_folder
        self.image_key = image_key
        if target_magnifications is not None:
            target_magnifications.sort(reverse=True)
        self.target_magnifications = target_magnifications if target_magnifications else None
        self.add_lowest_magnification = add_lowest_magnification
        self.replace_ampersand_by = replace_ampersand_by
        self.default_base_objective_power = default_base_objective_power
        self.wsi_reader = WSIReader(backend=WSIBackend.OPENSLIDE)
        self.compression = compression
        self.tile_size = tile_size
        self.min_file_size = min_file_size
        self.verbose = verbose

    def get_tiff_path(self, src_path: Path) -> Path:
        """Returns the path to the tiff file that will be created from the src file. The tiff file is saved in the
        output_folder with the same name as the src file but with the tiff extension. Ampersands are replaced by the
        replace_ampersand_by string.

        :param src_path: The path to the src file.
        :return: The path to the tiff file that will be created from the src file.
        """
        tiff_filename = src_path.with_suffix(TIFF_EXTENSION).name
        tiff_filename = tiff_filename.replace(AMPERSAND, self.replace_ampersand_by)
        return self.output_folder / tiff_filename

    def _get_base_objective_power(self, wsi_obj: OpenSlide) -> float:
        """Returns the objective power of the wsi. The objective power is extracted from the wsi properties. If the
        objective power is not found in the wsi properties, the base_objective_power is used instead.

        :param wsi_obj: The wsi object in openslide format
        :raises ValueError: Raises an error if the objective power is not found in the wsi properties and
            base_objective_power is None
        :return: The base objective power of the wsi
        """
        base_objective_power = wsi_obj.properties.get(self.OBJECTIVE_POWER_KEY, self.default_base_objective_power)
        if base_objective_power is None:
            raise ValueError(
                f"Could not find {self.OBJECTIVE_POWER_KEY} in wsi properties. Please specify a default value for "
                "default_base_objective_power."
            )
        return float(base_objective_power)

    def _get_target_level(self, wsi_obj: OpenSlide, target_magnification: float) -> int:
        """Returns the level of the wsi pyramid that is equal to the target magnification.

        :param wsi_obj: the wsi object in openslide format
        :param target_magnification: the target magnification e.g. 10x
        """
        base_objective_power = self._get_base_objective_power(wsi_obj)
        for level_idx, level_downsample in enumerate(wsi_obj.level_downsamples):
            objective_power = base_objective_power / level_downsample
            if math.isclose(objective_power, target_magnification, rel_tol=1e-3):
                return level_idx
        raise ValueError(f"Target magnification {target_magnification} not found")

    def _get_highest_level(self, wsi_obj: OpenSlide) -> int:
        """Returns the highest resolution level of the wsi pyramid."""
        return len(wsi_obj.level_downsamples) - 1

    def get_target_levels(self, wsi_obj: Any) -> List[int]:
        """Returns the levels of the wsi pyramid that will be saved in the tiff file. If target_magnifications is not
        None, we return the levels that correspond to the target magnifications. If add_lowest_magnification is True, we
        add the highest resolution level of the wsi pyramid if it is not already in the target levels.
        Otherwise, we return all levels.

        :param wsi_obj: The wsi object in openslide format
        :return: A list of levels that will be saved in the tiff file
        """
        if self.target_magnifications is not None:
            target_levels = [self._get_target_level(wsi_obj, mag) for mag in self.target_magnifications]
            if self.add_lowest_magnification:
                highest_level = self._get_highest_level(wsi_obj)
                if highest_level not in target_levels:
                    target_levels.append(highest_level)
            return target_levels
        return [level for level in range(len(wsi_obj.level_downsamples))]

    def get_level_data(self, wsi_obj: OpenSlide, level: int) -> np.ndarray:
        """Returns data from a given level of the wsi pyramid. The data is returned in HWC format as expected by the
        tiffwriter. The data must be in HWC format otherwise the compression will not work properly.

        :param wsi_obj: The wsi object in openslide format
        :param level: The level at which the data is extracted from the wsi pyramid
        :return: A numpy array of shape (H, W, C) where H is the height, W is the width and C is the number of channels
        """
        level_data, _ = self.wsi_reader.get_data(wsi_obj, level=level)
        level_data = level_data.transpose(1, 2, 0)
        self.validate_level_data(level_data)
        return level_data

    def validate_level_data(self, level_data: np.ndarray) -> None:
        """Makes sure that the level data is in HWC format. If the level data is not in HWC format, an error is raised.
        This sanity check is necessary to avoid compression issues.
        """
        if level_data.shape[2] != 3:
            raise ValueError(
                f"Expected 3 channels but got {level_data.shape[2]}. Maybe the image is in channel first format? Try "
                "transposing the image."
            )

    def get_tiffwriter_options(self, wsi_obj: OpenSlide) -> Dict[str, Any]:
        """Returns the options that will be passed to the tiffwriter. The options are extracted from the wsi properties
        and will be written as tags in the tiff file.

        :param wsi_obj: The wsi object in openslide format
        :raises ValueError: Raises an error if the resolution unit is not in centimeters
        :return: A dictionary of options that will be passed to the tiffwriter
        """
        resolution_unit = wsi_obj.properties[self.RESOLUTION_UNIT_KEY]

        if resolution_unit != self.RESOLUTION_UNIT:
            logger.info(f"Resolution unit is not in {self.RESOLUTION_UNIT}: {resolution_unit}")

        options = dict(
            software=self.SOFTWARE,
            metadata={'axes': 'YXC'},
            photometric=PHOTOMETRIC.RGB,
            resolutionunit=resolution_unit,
            compression=self.compression,
            tile=(self.tile_size, self.tile_size),
        )
        return options

    def get_px_per_cm_resolution_at_level(self, wsi_obj: OpenSlide, level: int) -> Tuple[float, float]:
        """Returns the resolution of the wsi at a given level in pixels per centimeter.

        :param wsi_obj: The wsi object in openslide format
        :param level: The level at which the resolution is calculated
        :return: A tuple of floats (x_resolution, y_resolution)
        """
        um_per_cm = 10000
        try:
            um_per_px = self.wsi_reader.get_mpp(wsi_obj, level=level)
        except KeyError:
            um_per_px = (wsi_obj.properties["openslide.mpp-x"], wsi_obj.properties["openslide.mpp-y"])
        px_per_cm = (um_per_cm / um_per_px[0], um_per_cm / um_per_px[1])
        return px_per_cm

    def convert_wsi(self, src_path: Path, tiff_path: Path) -> None:
        """Converts a single wsi file from a src format to tiff format. The tiff file is saved in the tiff_path. If the
        original src file does not have the target resolution, we skip the wsi and return None.

        :param src_path: The path to the src wsi file
        :param tiff_path: The path to the tiff file
        """

        wsi_obj = self.wsi_reader.read(src_path)

        try:
            levels = self.get_target_levels(wsi_obj)
        except ValueError as e:
            logger.warning(f"Skipping {src_path} because {e}")
            return

        options = self.get_tiffwriter_options(wsi_obj)

        with TiffWriter(tiff_path, bigtiff=True) as tif:
            for i, level in enumerate(levels):
                level_data = self.get_level_data(wsi_obj, level)
                self.validate_level_data(level_data)
                resolution = self.get_px_per_cm_resolution_at_level(wsi_obj, level)
                # the subfiletype parameter is a bitfield that determines if the wsi_level is a reduced version of
                # another image. level 0 (i.e. i=0) is the full resolution image in the pyramid.
                tif.write(level_data, resolution=resolution, subfiletype=int(i > 0), **options)

    def __call__(self, data: Dict) -> Dict:
        src_path = Path(data[self.image_key])
        tiff_path = self.get_tiff_path(src_path)
        # if the tiff file does not exist or if it exists but is empty, we convert the wsi to tiff
        if not tiff_path.exists() or (tiff_path.exists() and tiff_path.stat().st_size <= self.min_file_size):
            self.convert_wsi(src_path, tiff_path)
        if self.verbose:
            logger.info(f"Converted {src_path} to {tiff_path}")
            logger.info(f"Source file size {src_path.stat().st_size / 1e6:.2f} MB")
            logger.info(f"Tiff file size {tiff_path.stat().st_size / 1e6:.2f} MB")
        return data



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import logging
import torch
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Dict, Generic, List, Optional, Sequence, Tuple, TypeVar, Union

from pytorch_lightning import LightningDataModule
from pytorch_lightning.overrides.distributed import UnrepeatedDistributedSampler
from torch.utils.data import DataLoader, DistributedSampler
from health_cpath.preprocessing.loading import LoadingParams

from health_ml.utils.bag_utils import BagDataset, multibag_collate
from health_ml.utils.common_utils import _create_generator

from health_cpath.utils.wsi_utils import TilingParams, image_collate
from health_cpath.models.transforms import LoadTilesBatchd
from health_cpath.datasets.base_dataset import SlidesDataset, TilesDataset
from health_cpath.utils.naming import ModelKey

from monai.data.dataset import CacheDataset, Dataset, PersistentDataset
from monai.transforms import Compose


_SlidesOrTilesDataset = TypeVar("_SlidesOrTilesDataset", SlidesDataset, TilesDataset)


class CacheMode(Enum):
    NONE = "none"
    MEMORY = "memory"
    DISK = "disk"


class CacheLocation(Enum):
    NONE = "none"
    CPU = "cpu"
    SAME = "same"


class HistoDataModule(LightningDataModule, Generic[_SlidesOrTilesDataset]):
    """Base class to load a histopathology dataset as train, val, test sets"""

    def __init__(
        self,
        root_path: Path,
        batch_size: int = 1,
        batch_size_inf: Optional[int] = None,
        max_bag_size: int = 0,
        max_bag_size_inf: int = 0,
        seed: Optional[int] = None,
        transforms_dict: Optional[Dict[ModelKey, Union[Callable, None]]] = None,
        crossval_count: int = 0,
        crossval_index: int = 0,
        pl_replace_sampler_ddp: bool = True,
        dataloader_kwargs: Optional[Dict[str, Any]] = None,
        dataframe_kwargs: Optional[Dict[str, Any]] = None,
        class_weights: Optional[torch.Tensor] = None,
    ) -> None:
        """
        :param root_path: Root directory of the source dataset.
        :param batch_size: Number of slides to load per batch.
        :param batch_size_inf: Number of slides to load per batch during inference. If None, use batch_size.
        :param max_bag_size: Upper bound on number of tiles in each loaded bag during training stage. If 0 (default),
        will return all samples in each bag. If > 0 , bags larger than `max_bag_size` will yield
        random subsets of instances. For SlideDataModule, this parameter is used in Rand/GridPatchd Transform to set the
        num_patches used for tiling on the fly at training time.
        :param max_bag_size_inf: Upper bound on number of tiles in each loaded bag during validation and test stages.
        If 0 (default), will return all samples in each bag. If > 0 , bags larger than `max_bag_size_inf` will yield
        random subsets of instances. For SlideDataModule, this parameter is used in Rand/GridPatchd Transform to set the
        num_patches used for tiling on the fly at validation and test time.
        :param seed: pseudorandom number generator seed to use for shuffling instances and bags. Note that randomness in
        train/val/test splits is handled independently in `get_splits()`. (default: `None`)
        :param transforms_dict: A dictionary that contains transform, or a composition of transforms using
        `monai.transforms.Compose`, to apply to the source dataset at training, validation and testing time.
        By default (`None`).
        :param crossval_count: Number of folds to perform.
        :param crossval_index: Index of the cross validation split to be performed.
        :param pl_replace_sampler_ddp: If True, replace the sampler with a DistributedSampler when using DDP.
        :param dataloader_kwargs: Additional keyword arguments for the training, validation, and test dataloaders.
        :param dataframe_kwargs: Keyword arguments to pass to `pd.read_csv()` when loading the dataset CSV.
        :param class_weights: Class weights to use for the dataset. If None, will compute them from the dataset.
        """

        batch_size_inf = batch_size_inf or batch_size
        super().__init__()

        self.root_path = root_path
        self.transforms_dict = transforms_dict
        self.batch_sizes = {ModelKey.TRAIN: batch_size, ModelKey.VAL: batch_size_inf, ModelKey.TEST: batch_size_inf}
        self.bag_sizes = {ModelKey.TRAIN: max_bag_size, ModelKey.VAL: max_bag_size_inf, ModelKey.TEST: max_bag_size_inf}
        self.crossval_count = crossval_count
        self.crossval_index = crossval_index
        self.pl_replace_sampler_ddp = pl_replace_sampler_ddp
        self.train_dataset: _SlidesOrTilesDataset
        self.val_dataset: _SlidesOrTilesDataset
        self.test_dataset: _SlidesOrTilesDataset
        self.dataframe_kwargs = dataframe_kwargs or {}
        self.train_dataset, self.val_dataset, self.test_dataset = self.get_splits()
        self.class_weights = class_weights if class_weights is not None else self.train_dataset.get_class_weights()
        self.seed = seed
        self.dataloader_kwargs = dataloader_kwargs or {}

    def get_splits(self) -> Tuple[_SlidesOrTilesDataset, _SlidesOrTilesDataset, _SlidesOrTilesDataset]:
        """Create the training, validation, and test datasets"""
        raise NotImplementedError

    def _get_dataloader(
        self, dataset: _SlidesOrTilesDataset, stage: ModelKey, shuffle: bool, **dataloader_kwargs: Any
    ) -> DataLoader:
        raise NotImplementedError

    def _get_ddp_sampler(self, dataset: Dataset, stage: ModelKey) -> Optional[DistributedSampler]:
        is_distributed = torch.distributed.is_initialized() and torch.distributed.get_world_size() > 1
        if is_distributed and not self.pl_replace_sampler_ddp:
            assert self.seed is not None, "seed must be set when using distributed training for reproducibility"
            if stage == ModelKey.TRAIN:
                logging.info("pl_replace_sampler_ddp is False, setting DistributedSampler for training dataloader.")
                return DistributedSampler(dataset, shuffle=True, seed=self.seed)
            else:
                logging.info(
                    "pl_replace_sampler_ddp is False, setting UnrepeatedDistributedSampler for validation and "
                    "test dataloaders. This will ensure that each process gets a unique set of samples. "
                    "If you want to use DistributedSampler, set pl_replace_sampler_ddp to True."
                )
                return UnrepeatedDistributedSampler(dataset, shuffle=False, seed=self.seed)
        return None

    def train_dataloader(self) -> DataLoader:
        return self._get_dataloader(
            self.train_dataset, shuffle=True, stage=ModelKey.TRAIN, **self.dataloader_kwargs  # type: ignore
        )

    def val_dataloader(self) -> DataLoader:
        return self._get_dataloader(
            self.val_dataset, shuffle=False, stage=ModelKey.VAL, **self.dataloader_kwargs  # type: ignore
        )

    def test_dataloader(self) -> DataLoader:
        return self._get_dataloader(
            self.test_dataset, shuffle=False, stage=ModelKey.TEST, **self.dataloader_kwargs  # type: ignore
        )


class TilesDataModule(HistoDataModule[TilesDataset]):
    """Base class to load the tiles of a dataset as train, val, test sets. Note that tiles are always shuffled by
    default. This means that we sample a random subset of tiles from each bag at each epoch. This is different from
    slides shuffling that is switched on during training time only. This is done to avoid overfitting to the order of
    the tiles in each bag.
    """

    def __init__(
        self,
        cache_mode: CacheMode = CacheMode.NONE,
        precache_location: CacheLocation = CacheLocation.NONE,
        cache_dir: Optional[Path] = None,
        **kwargs: Any,
    ) -> None:
        """
        :param cache_mode: The type of caching to perform, i.e. whether the results of all
        transforms up to the first randomised one should be computed only once and reused in
        subsequent iterations:
          - `MEMORY`: MONAI CacheDataset is used, the entire transformed dataset is kept in memory for fastest access;
          - `DISK`: MONAI PersistentDataset is used, each transformed sample is saved to disk and loaded on-demand;
          - `NONE` (default): standard MONAI dataset is used, no caching is performed.
        :param precache_location: Whether to pre-cache the entire transformed dataset upfront and save
        it to disk. This is done once in `prepare_data()` only on the local rank-0 process, so
        multiple processes can afterwards access the same cache without contention in DDP settings.
        This parameter also allows us to choose if the cache will be re-loaded into CPU or GPU memory:
          - `NONE (default)`: no pre-cache is performed;
          - `CPU`: each transformed sample is saved to disk and, if cache_mode is `MEMORY`, reloaded into CPU;
          - `SAME`: each transformed sample is saved to disk and, if cache_mode is `MEMORY`, reloaded on the same
          device it was saved from;
        If cache_mode is `DISK` precache_location `CPU` and `GPU` are equivalent.
        :param cache_dir: The directory onto which to cache data if caching is enabled.
        """
        if precache_location is not CacheLocation.NONE and cache_mode is CacheMode.NONE:
            raise ValueError("Can only pre-cache if caching is enabled")
        if precache_location is not CacheLocation.NONE and cache_dir is None:
            raise ValueError("A cache directory is required for pre-caching")
        if cache_mode is CacheMode.DISK and cache_dir is None:
            raise ValueError("A cache directory is required for on-disk caching")

        self.cache_mode = cache_mode
        self.precache_location = precache_location
        self.cache_dir = cache_dir

        super().__init__(**kwargs)

    def prepare_data(self) -> None:
        if self.precache_location != CacheLocation.NONE:
            self._load_dataset(self.train_dataset, stage=ModelKey.TRAIN, shuffle=True)
            self._load_dataset(self.val_dataset, stage=ModelKey.VAL, shuffle=True)
            self._load_dataset(self.test_dataset, stage=ModelKey.TEST, shuffle=True)

    def _dataset_pickle_path(self, stage: str) -> Optional[Path]:
        if self.cache_dir is None or self.cache_mode == CacheMode.NONE:
            return None
        return self.cache_dir / f"{stage}_dataset.pt"

    def _get_transformed_dataset(self, dataset: Dataset, transform: Union[Sequence[Callable], Callable]) -> Dataset:
        if self.cache_mode is CacheMode.MEMORY:
            dataset = CacheDataset(dataset, transform, num_workers=1)  # type: ignore
        elif self.cache_mode is CacheMode.DISK:
            dataset = PersistentDataset(dataset, transform, cache_dir=self.cache_dir)  # type: ignore
            if self.precache_location != CacheLocation.NONE:
                import tqdm  # TODO: Make optional

                for i in tqdm.trange(len(dataset), desc="Loading dataset"):
                    dataset[i]  # empty loop to pre-compute all transformed samples
        else:
            dataset = Dataset(dataset, transform)  # type: ignore
        return dataset

    def _load_dataset(self, tiles_dataset: TilesDataset, stage: ModelKey, shuffle: bool) -> Dataset:
        dataset_pickle_path = self._dataset_pickle_path(stage)

        if dataset_pickle_path and dataset_pickle_path.is_file():
            if self.precache_location == CacheLocation.CPU:
                memory_location = torch.device("cpu")
                print(f"Loading dataset from {dataset_pickle_path} into {memory_location}")
            else:
                # by default torch.load will reload on the same device it was saved from
                memory_location = None  # type: ignore

            with dataset_pickle_path.open("rb") as f:
                return torch.load(f, map_location=memory_location)

        generator = _create_generator(self.seed)

        bag_dataset = BagDataset(
            tiles_dataset,  # type: ignore
            bag_ids=tiles_dataset.slide_ids,
            max_bag_size=self.bag_sizes[stage],
            shuffle_samples=True,
            generator=generator,
        )
        if self.transforms_dict and self.transforms_dict[stage]:
            transform = self.transforms_dict[stage]
        else:
            transform = LoadTilesBatchd(tiles_dataset.IMAGE_COLUMN)

        # Save and restore PRNG state for consistency across (pre-)caching options
        generator_state = generator.get_state()
        transformed_bag_dataset = self._get_transformed_dataset(bag_dataset, transform)  # type: ignore
        generator.set_state(generator_state)

        # Dataset is saved if cache_dir is True, regardless of CacheMode
        if dataset_pickle_path:
            dataset_pickle_path.parent.mkdir(parents=True, exist_ok=True)
            with dataset_pickle_path.open("wb") as f:
                torch.save(transformed_bag_dataset, f)

        return transformed_bag_dataset

    def _get_dataloader(
        self, dataset: TilesDataset, stage: ModelKey, shuffle: bool, **dataloader_kwargs: Any
    ) -> DataLoader:
        transformed_bag_dataset = self._load_dataset(dataset, stage=stage, shuffle=shuffle)
        bag_dataset: BagDataset = transformed_bag_dataset.data  # type: ignore
        generator = bag_dataset.bag_sampler.generator
        sampler = self._get_ddp_sampler(transformed_bag_dataset, stage)
        return DataLoader(
            transformed_bag_dataset,
            batch_size=self.batch_sizes[stage],
            collate_fn=multibag_collate,
            sampler=sampler,
            # sampler option is mutually exclusive with shuffle
            shuffle=shuffle if sampler is None else None,  # type: ignore
            generator=generator,
            **dataloader_kwargs,
        )


class SlidesDataModule(HistoDataModule[SlidesDataset]):
    """
    Base class to load the slides of a dataset as train, val, test sets. The slide data module performs tiling on the
    fly by default. One can specify the tiling strategies (background removal, overlapping tiles, padding, ...) through
    the class parameters.
    """

    def __init__(
        self,
        loading_params: LoadingParams,
        tiling_params: TilingParams,
        **kwargs: Any,
    ) -> None:
        """
        :param tiling_params: the tiling on the fly parameters.
        :param loading_params: the loading parameters.
        :param kwargs: additional parameters to pass to the parent class HistoDataModule
        """
        super().__init__(**kwargs)
        self.tiling_params = tiling_params
        self.loading_params = loading_params

    def _load_dataset(self, slides_dataset: SlidesDataset, stage: ModelKey) -> Dataset:
        base_transform = Compose(self.get_tiling_transforms(stage=stage))
        if self.transforms_dict and self.transforms_dict[stage]:
            transforms = Compose([base_transform, self.transforms_dict[stage]]).flatten()
        else:
            transforms = base_transform
        # The tiling transform is randomized. Make them deterministic. This call needs to be
        # done on the final Compose, not at the level of the individual randomized transforms.
        transforms.set_random_state(seed=self.seed)
        return Dataset(slides_dataset, transforms)

    def _get_dataloader(
        self, dataset: SlidesDataset, stage: ModelKey, shuffle: bool, **dataloader_kwargs: Any
    ) -> DataLoader:
        transformed_slides_dataset = self._load_dataset(dataset, stage)
        generator = _create_generator(self.seed)
        sampler = self._get_ddp_sampler(transformed_slides_dataset, stage)
        return DataLoader(
            transformed_slides_dataset,
            batch_size=self.batch_sizes[stage],
            collate_fn=image_collate,
            sampler=sampler,
            # sampler option is mutually exclusive with shuffle
            shuffle=shuffle if not sampler else None,  # type: ignore
            generator=generator,
            **dataloader_kwargs,
        )

    def get_tiling_transforms(self, stage: ModelKey) -> List[Callable]:
        """Returns the list of transforms to apply to the dataset to perform tiling on the fly. The transforms are
        applied in the order they are returned by this method. To add additional transforms, override this method."""
        return [
            self.loading_params.get_load_roid_transform(),
            self.tiling_params.get_tiling_transform(bag_size=self.bag_sizes[stage], stage=stage),
            self.tiling_params.get_extract_coordinates_transform(),
            self.tiling_params.get_split_transform(),
        ]



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

"""This script is specific to PANDA and is kept only for retrocompatibility.
`create_tiles_dataset.py` is the new supported way to process slide datasets.
"""
import functools
import sys
import shutil
from pathlib import Path
from argparse import ArgumentParser
from typing import Tuple, Union, List

import numpy as np
from monai.data import Dataset
from tqdm import tqdm
from health_ml.utils.box_utils import Box

from health_cpath.preprocessing import tiling
from health_cpath.utils.naming import SlideKey, TileKey
from health_cpath.datasets.panda_dataset import PandaDataset
from health_cpath.preprocessing.loading import LoadMaskROId, WSIBackend
from health_cpath.preprocessing.create_tiles_dataset import get_tile_id, save_image, merge_dataset_csv_files

CSV_COLUMNS = (
    'slide_id',
    'tile_id',
    'image',
    'mask',
    'left',
    'top',
    'right',
    'bottom',
    'occupancy',
    'data_provider',
    'slide_isup_grade',
    'slide_gleason_score',
    'num_discarded',
)
TMP_SUFFIX = "_tmp"


def select_tile(
    mask_tile: np.ndarray, occupancy_threshold: float
) -> Union[Tuple[bool, float], Tuple[np.ndarray, np.ndarray]]:
    if occupancy_threshold < 0.0 or occupancy_threshold > 1.0:
        raise ValueError("Tile occupancy threshold must be between 0 and 1")
    # mask_tile has shape (N, C, H, W)
    foreground_mask = mask_tile > 0
    occupancy = foreground_mask.mean(axis=(-2, -1))
    selected = occupancy >= occupancy_threshold  # if the threshold is 0, all tiles should be selected
    # selected has shape (N, 1)
    return selected[:, 0], occupancy[:, 0]


def generate_tiles(
    sample: dict, tile_size: int, occupancy_threshold: float
) -> Tuple[np.ndarray, np.ndarray, List[Box], np.ndarray, int]:
    image_tiles, tile_locations = tiling.tile_array_2d(sample['image'], tile_size=tile_size, constant_values=255)
    assert tile_locations.ndim == 2
    mask_tiles, _ = tiling.tile_array_2d(sample['mask'], tile_size=tile_size, constant_values=0)

    selected: np.ndarray
    occupancies: np.ndarray
    selected, occupancies = select_tile(mask_tiles, occupancy_threshold)  # type: ignore
    num_selected = selected.sum()
    num_tiles = len(image_tiles)
    num_discarded = num_tiles - num_selected
    percentage_discarded = 100 * num_discarded / num_tiles
    print(f"Discarded {num_discarded}/{num_tiles} tiles ({percentage_discarded:.2f} %)")

    image_tiles = image_tiles[selected]
    mask_tiles = mask_tiles[selected]
    tile_locations = tile_locations[selected]
    occupancies = occupancies[selected]

    top, left = sample['location']
    offset = np.array((left, top))

    abs_tile_locations = (sample['scale'] * tile_locations + offset).astype(int)
    tile_size_scaled = int(tile_size * sample['scale'])
    tile_boxes = [Box(x, y, tile_size_scaled, tile_size_scaled) for x, y in abs_tile_locations]

    return image_tiles, mask_tiles, tile_boxes, occupancies, num_discarded


# TODO refactor this to separate metadata identification from saving. We might want the metadata
# even if the saving fails
def save_tile(sample: dict, image_tile: np.ndarray, mask_tile: np.ndarray, tile_box: Box, output_dir: Path) -> dict:
    slide_id = sample[SlideKey.SLIDE_ID]
    tile_id = get_tile_id(slide_id, tile_box)
    image_tile_filename = f"train_images/{tile_id}.png"
    mask_tile_filename = f"train_label_masks/{tile_id}_mask.png"

    save_image(image_tile, output_dir / image_tile_filename)
    save_image(mask_tile, output_dir / mask_tile_filename)

    slide_metadata = sample[SlideKey.METADATA]
    tile_metadata = {
        TileKey.SLIDE_ID.value: slide_id,
        TileKey.TILE_ID.value: tile_id,
        TileKey.IMAGE.value: image_tile_filename,
        TileKey.MASK.value: mask_tile_filename,
        TileKey.TILE_LEFT.value: tile_box.x,
        TileKey.TILE_TOP.value: tile_box.y,
        TileKey.TILE_RIGHT.value: tile_box.x + tile_box.w,
        TileKey.TILE_BOTTOM.value: tile_box.y + tile_box.h,
        'data_provider': slide_metadata['data_provider'],
        'slide_isup_grade': slide_metadata['isup_grade'],
        'slide_gleason_score': slide_metadata['gleason_score'],
    }
    return tile_metadata


def process_slide(
    sample: dict,
    level: int,
    margin: int,
    tile_size: int,
    occupancy_threshold: int,
    output_dir: Path,
    tile_progress: bool = False,
    filter_slide: str = '',
) -> None:
    slide_id = sample[SlideKey.SLIDE_ID]
    if filter_slide not in slide_id:
        return
    slide_dir: Path = output_dir / (slide_id + "/")
    print(f">>> Slide dir {slide_dir}")
    if slide_dir.exists():  # already processed slide - skip
        print(f">>> Skipping {slide_dir} - already processed")
        return
    else:
        mask_key = SlideKey.MASK  # it should be read from the dataset attribute instead, but we assume it's the same
        mask_path = Path(sample[mask_key])
        if not mask_path.is_file():
            print(f'Mask for slide {slide_id} not found')
            return

        slide_dir.mkdir(parents=True)

        dataset_csv_path = slide_dir / "dataset.csv"
        dataset_csv_file = dataset_csv_path.open('w')
        dataset_csv_file.write(','.join(CSV_COLUMNS) + '\n')  # write CSV header

        print(f"Loading slide {slide_id} ...")
        loader = LoadMaskROId(backend=WSIBackend.CUCIM, level=level, margin=margin)
        try:
            sample = loader(sample)  # load 'image' and 'mask' from disk
            failed = False
        except RuntimeError as e:  # happens when masks are empty
            print(f'Error loading slide {slide_id}, maybe due to an empty mask:\n{e}')
            failed = True

        if failed:
            print(f'Error loading slide {slide_id}')
            dataset_csv_file.close()
            shutil.rmtree(slide_dir)
            return

        print(f"Tiling slide {slide_id} ...")
        image_tiles, mask_tiles, tile_boxes, occupancies, num_discarded = generate_tiles(
            sample, tile_size, occupancy_threshold
        )
        n_tiles = image_tiles.shape[0]

        for i in tqdm(range(n_tiles), f"Tiles ({slide_id[:6]}…)", unit="img", disable=not tile_progress):
            tile_metadata = save_tile(
                sample,
                image_tiles[i],
                mask_tiles[i],
                tile_boxes[i],
                slide_dir,
            )
            relative_slide_dir = Path(slide_dir.name)
            tile_metadata[TileKey.OCCUPANCY] = occupancies[i]
            tile_metadata[TileKey.IMAGE] = relative_slide_dir / tile_metadata[TileKey.IMAGE]
            tile_metadata[TileKey.MASK] = relative_slide_dir / tile_metadata[TileKey.MASK]
            tile_metadata[TileKey.NUM_DISCARDED] = num_discarded
            dataset_row = ','.join(str(tile_metadata[column]) for column in CSV_COLUMNS)
            dataset_csv_file.write(dataset_row + '\n')

        dataset_csv_file.close()


def main(
    panda_dir: Union[str, Path],
    root_output_dir: str,
    level: int,
    tile_size: int,
    margin: int,
    occupancy_threshold: float,
    parallel: bool = False,
    overwrite: bool = False,
    filter_slide: str = '',
) -> None:
    # Ignoring some types here because mypy is getting confused with the MONAI Dataset class
    # to select a subsample use keyword n_slides
    dataset = Dataset(PandaDataset(panda_dir))  # type: ignore
    output_dir = Path(root_output_dir)

    if overwrite and output_dir.exists():
        shutil.rmtree(output_dir)
    output_dir.mkdir(parents=True, exist_ok=not overwrite)

    print(f"Command: \"{' '.join(sys.argv)}\"")
    print(f"Creating dataset of level-{level} {tile_size}x{tile_size} PANDA tiles at: {output_dir}")

    func = functools.partial(
        process_slide,
        level=level,
        margin=margin,
        tile_size=tile_size,
        occupancy_threshold=occupancy_threshold,
        output_dir=output_dir,
        tile_progress=not parallel,
        filter_slide=filter_slide,
    )

    if parallel:
        import multiprocessing

        pool = multiprocessing.Pool()
        map_func = pool.imap_unordered  # type: ignore
    else:
        map_func = map  # type: ignore

    list(tqdm(map_func(func, dataset), desc="Slides", unit="img", total=len(dataset)))  # type: ignore

    if parallel:
        pool.close()

    print("Merging slide files in a single file")
    merge_dataset_csv_files(output_dir)


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument(
        "--panda-dir",
        type=str,
        default="/tmp/datasets/PANDA",
        help="Folder with the PANDA dataset. For example, '/tmp/datasets/PANDA'",
    )
    parser.add_argument("--output-dir", type=str, default="/datasetdrive/PANDA_20X_level_0_224")
    parser.add_argument(
        "--level",
        type=int,
        default=0,
    )
    parser.add_argument(
        "--tile-size",
        type=int,
        default=224,
    )
    parser.add_argument(
        "--margin",
        type=int,
        default=0,
    )
    parser.add_argument(
        "--min-occupancy",
        type=float,
        default=0.05,
    )
    parser.add_argument(
        "--no-parallel",
        action="store_false",
        default=True,
        dest="parallel",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        default=True,
    )
    parser.add_argument(
        "--filter-slide",
        type=str,
        default="",  # filtering for "b896" gives 4 slides, good for debugging
        help="Process only slides whose ID contain this substring. Useful for debugging",
    )
    args = parser.parse_args()

    main(
        panda_dir=args.panda_dir,
        root_output_dir=args.output_dir,
        level=args.level,
        tile_size=args.tile_size,
        margin=args.margin,
        occupancy_threshold=args.min_occupancy,
        parallel=args.parallel,
        overwrite=args.overwrite,
        filter_slide=args.filter_slide,
    )



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
import pandas as pd
from pathlib import Path
from typing import Any, Dict, Union, Optional
from health_cpath.datasets.base_dataset import SlidesDataset


class PandaColumns:
    SLIDE_ID = 'image_id'
    IMAGE = 'image'
    MASK = 'mask'
    METADATA = ('data_provider', 'isup_grade', 'gleason_score')


PANDA_CSV_FILENAME = "train.csv"


class PandaDataset(SlidesDataset):
    """Dataset class for loading files from the PANDA challenge dataset.

    Iterating over this dataset returns a dictionary following the `SlideKey` schema plus meta-data
    from the original dataset (`'data_provider'`, `'isup_grade'`, and `'gleason_score'`).

    Ref.: https://www.kaggle.com/c/prostate-cancer-grade-assessment/overview
    """

    def __init__(
        self,
        root: Union[str, Path],
        dataset_csv: Optional[Union[str, Path]] = None,
        dataset_df: Optional[pd.DataFrame] = None,
        label_column: str = "isup_grade",
        n_classes: int = 6,
        dataframe_kwargs: Dict[str, Any] = {},
    ) -> None:
        super().__init__(
            root,
            dataset_csv,
            dataset_df,
            validate_columns=False,
            label_column=label_column,
            n_classes=n_classes,
            dataframe_kwargs=dataframe_kwargs,
            slide_id_column=PandaColumns.SLIDE_ID,
            image_column=PandaColumns.IMAGE,
            mask_column=PandaColumns.MASK,
            metadata_columns=PandaColumns.METADATA,
            default_csv_filename=PANDA_CSV_FILENAME,
        )
        # PANDA CSV does not come with paths for image and mask files
        slide_ids = self.dataset_df.index
        self.dataset_df[self.image_column] = "train_images/" + slide_ids + ".tiff"
        self.dataset_df[self.mask_column] = "train_label_masks/" + slide_ids + "_mask.tiff"
        self.validate_columns()



#  -------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  -------------------------------------------------------------------------------------------

from pathlib import Path
from typing import Any, Dict, List, Sequence, Tuple

import dateutil.parser
import numpy as np
import pandas as pd
from azureml.core import Experiment, Run, Workspace

from health_ml.utils.common_utils import df_to_json
from health_azure.utils import (
    aggregate_hyperdrive_metrics,
    download_file_if_necessary,
    get_aml_run_from_run_id,
    get_tags_from_hyperdrive_run,
)
from health_cpath.utils.output_utils import (
    AML_LEGACY_TEST_OUTPUTS_CSV,
    AML_TEST_OUTPUTS_CSV,
    AML_VAL_OUTPUTS_CSV,
    validate_class_names,
)
from health_cpath.utils.naming import AMLMetricsJsonKey


def run_has_val_and_test_outputs(run: Run) -> bool:
    """Checks whether the given run has both validation and test outputs files.

    :param parent_run: The run whose outputs to check.
    :raises ValueError: If the run does not have the expected output file(s).
    :return: `True` if the run has validation and test outputs, `False` if it is a legacy run with
        only test outputs.
    """
    available_files: List[str] = run.get_file_names()

    if AML_VAL_OUTPUTS_CSV in available_files and AML_TEST_OUTPUTS_CSV in available_files:
        return True
    elif AML_LEGACY_TEST_OUTPUTS_CSV in available_files:
        return False
    else:
        raise ValueError(
            f"Run {run.display_name} ({run.id}) does not have the expected files "
            f"({AML_LEGACY_TEST_OUTPUTS_CSV} or both {AML_VAL_OUTPUTS_CSV} and "
            f"{AML_TEST_OUTPUTS_CSV}): {available_files}"
        )


def child_runs_have_val_and_test_outputs(parent_run: Run) -> bool:
    """Checks whether all child hyperdrive runs have both validation and test outputs files.

    :param parent_run: The parent Hyperdrive run.
    :raises ValueError: If any of the child runs does not have the expected output files, or if
        some of the child runs have both outputs and some have only test outputs.
    :return: `True` if all children have validation and test outputs, `False` if all children are
        legacy runs with only test outputs.
    """
    have_val_and_test_outputs = [run_has_val_and_test_outputs(child_run) for child_run in parent_run.get_children()]
    if all(have_val_and_test_outputs):
        return True
    elif not any(have_val_and_test_outputs):
        return False
    else:
        raise ValueError(
            f"Parent run {parent_run.display_name} ({parent_run.id}) has mixed children with legacy "
            "test-only outputs and with both validation and test outputs"
        )


def collect_hyperdrive_outputs(
    parent_run_id: str,
    download_dir: Path,
    aml_workspace: Workspace,
    hyperdrive_arg_name: str = "crossval_index",
    output_filename: str = "test_output.csv",
    overwrite: bool = False,
) -> Dict[int, pd.DataFrame]:
    """Fetch output CSV files from Hyperdrive child runs as dataframes.

    Will only download the CSV files if they do not already exist locally.

    :param parent_run_id: Azure ML run ID for the parent Hyperdrive run.
    :param download_dir: Base directory where to download the CSV files. A new sub-directory will
        be created for each child run (e.g. `<download_dir>/<hyperdrive_arg_name>/*.csv`).
    :param aml_workspace: Azure ML workspace in which the runs were executed.
    :param hyperdrive_arg_name: Name of the Hyperdrive argument used for indexing the child runs.
    :param output_filename: Filename of the output CSVs to download.
    :param overwrite: Whether to force the download even if each file already exists locally.
    :return: A dictionary of dataframes with the sorted hyperdrive_arg_name indices as keys.
    """
    parent_run = get_aml_run_from_run_id(parent_run_id, aml_workspace)

    all_outputs_dfs = {}
    for child_run in parent_run.get_children():
        child_run_index = get_tags_from_hyperdrive_run(child_run, hyperdrive_arg_name)
        if child_run_index is None:
            raise ValueError(f"Child run expected to have the tag '{hyperdrive_arg_name}'")
        child_dir = download_dir / str(child_run_index)
        try:
            child_csv = download_file_if_necessary(
                child_run, output_filename, child_dir / output_filename, overwrite=overwrite
            )
            all_outputs_dfs[child_run_index] = pd.read_csv(child_csv)
        except Exception as e:
            print(f"Failed to download {output_filename} for run {child_run.id}: {e}")
    return dict(sorted(all_outputs_dfs.items()))  # type: ignore


def download_hyperdrive_metrics_if_required(
    parent_run_id: str,
    download_dir: Path,
    aml_workspace: Workspace,
    hyperdrive_arg_name: str = "crossval_index",
    overwrite: bool = False,
) -> Path:
    """Fetch metrics logged to Azure ML from hyperdrive runs.

    Will only download the metrics if they do not already exist locally, as this can take several
    seconds for each child run.

    :param parent_run_id: Azure ML run ID for the parent Hyperdrive run.
    :param download_dir: Directory where to save the downloaded metrics as `aml_metrics.json`.
    :param aml_workspace: Azure ML workspace in which the runs were executed.
    :param hyperdrive_arg_name: Name of the Hyperdrive argument used for indexing the child runs.
    :param overwrite: Whether to force the download even if metrics are already saved locally.
    :return: The path of the downloaded json file.
    """
    metrics_json = download_dir / "aml_metrics.json"
    if not overwrite and metrics_json.is_file():
        print(f"AML metrics file already exists at {metrics_json}")
    else:
        metrics_df = aggregate_hyperdrive_metrics(
            run_id=parent_run_id, child_run_arg_name=hyperdrive_arg_name, aml_workspace=aml_workspace
        )
        metrics_json.parent.mkdir(parents=True, exist_ok=True)
        print(f"Writing AML metrics file to {metrics_json}")
        df_to_json(metrics_df, metrics_json)
    return metrics_json


def collect_hyperdrive_metrics(metrics_json: Path) -> pd.DataFrame:
    """
    Collect the hyperdrive metrics from the downloaded metrics json file in a dataframe.
    :param metrics_json: Path of the downloaded metrics file `aml_metrics.json`.
    :return: A dataframe in the format returned by :py:func:`~health_azure.aggregate_hyperdrive_metrics()`.
    """
    metrics_df = pd.read_json(metrics_json).sort_index(axis='columns')
    return metrics_df


def get_hyperdrive_metrics_table(metrics_df: pd.DataFrame, metrics_list: Sequence[str]) -> pd.DataFrame:
    """Format raw hyperdrive metrics into a table with a summary "Mean ± Std" column.

    Note that this function only supports scalar metrics. To format metrics that are logged
    throughout training, you should call :py:func:`get_best_epoch_metrics()` first.

    :param metrics_df: Metrics dataframe, as returned by :py:func:`collect_hyperdrive_metrics()` and
        :py:func:`~health_azure.aggregate_hyperdrive_metrics()`.
    :param metrics_list: The list of metrics to include in the table.
    :return: A dataframe with the values of the selected metrics formatted as strings, including a
        header and a summary column.
    """
    header = ["Metric"] + [f"Child {k}" for k in metrics_df.columns] + ["Mean ± Std"]
    metrics_rows = []
    for metric in metrics_list:
        values: pd.Series = metrics_df.loc[metric]
        mean = values.mean()
        std = values.std()
        round_values: List[str] = [f"{v:.3f}" if v is not None else str(np.nan) for v in values]
        agg_values: List[str] = [f"{mean:.3f} ± {std:.3f}"]
        row = [metric] + round_values + agg_values
        metrics_rows.append(row)
    table = pd.DataFrame(metrics_rows, columns=header).set_index(header[0])
    return table


def get_best_epochs(
    metrics_df: pd.DataFrame, primary_metric: str, max_epochs_dict: Dict[int, int], maximise: bool = True
) -> Dict[int, Any]:
    """Determine the best epoch for each hyperdrive child run based on a given metric.

    The returned epoch indices are relative to the logging frequency of the chosen metric, i.e.
    should not be mixed between pipeline stages that log metrics at different epoch intervals.

    :param metrics_df: Metrics dataframe, as returned by :py:func:`collect_hyperdrive_metrics()` and
        :py:func:`~health_azure.aggregate_hyperdrive_metrics()`.
    :param primary_metric: Name of the reference metric to optimise.
    :max_epochs_dict: A dictionary of the maximum number of epochs in each cross-validation round.
    :param maximise: Whether the given metric should be maximised (minimised if `False`).
    :return: Dictionary mapping each hyperdrive child index to its best epoch.
    """
    best_epochs: Dict[int, Any] = {}
    for child_index in metrics_df.columns:
        primary_metric_list = metrics_df[child_index][primary_metric]
        if primary_metric_list is not None:
            # If extra validation epoch was logged (N+1), return only the first N elements
            primary_metric_list = (
                primary_metric_list[:-1]
                if (len(primary_metric_list) == max_epochs_dict[child_index] + 1)
                else primary_metric_list
            )
            best_epochs[child_index] = int(
                np.argmax(primary_metric_list) if maximise else np.argmin(primary_metric_list)
            )
        else:
            best_epochs[child_index] = None
    return best_epochs


def get_best_epoch_metrics(
    metrics_df: pd.DataFrame, metrics_list: Sequence[str], best_epochs: Dict[int, Any]
) -> pd.DataFrame:
    """Extract the values of the selected hyperdrive metrics at the given best epochs.

    The `best_epoch` indices are relative to the logging frequency of the chosen primary metric,
    i.e. the metrics in `metrics_list` must have been logged at the same epoch intervals.

    :param metrics_df: Metrics dataframe, as returned by :py:func:`collect_hyperdrive_metrics()` and
        :py:func:`~health_azure.aggregate_hyperdrive_metrics()`.
    :param metrics_list: Names of the metrics to index by the best epoch indices provided. Their
        values in `metrics_df` should be lists.
    :param best_epochs: Dictionary of hyperdrive child runs indices to best epochs, as returned by
        :py:func:`get_best_epochs()`.
    :return: Dataframe with the same columns as `metrics_df` and rows specified by `metrics_list`,
        containing only scalar values.
    """
    best_metrics = [
        metrics_df.loc[metrics_list, k].apply(lambda values: values[epoch])
        if epoch is not None
        else metrics_df.loc[metrics_list, k]
        for k, epoch in best_epochs.items()
    ]
    best_metrics_df = pd.DataFrame(best_metrics).T
    return best_metrics_df


def get_formatted_run_info(parent_run: Run) -> str:
    """Format Azure ML hyperdrive run information as HTML.

    Includes details of the parent and child runs, as well as submission information.

    :param parent_run: Parent Hyperdrive Azure ML run object.
    :return: Formatted HTML string.
    """

    def format_experiment(experiment: Experiment) -> str:
        return f"<a href={experiment.get_portal_url()}>{experiment.name}</a>"

    def format_run(run: Run) -> str:
        return f"<a href={run.get_portal_url()}>{run.display_name}</a> ({run.id}, {run.get_status()})"

    def format_submission_info(run: Run) -> str:
        details = run.get_details()
        start_time = dateutil.parser.parse(details['startTimeUtc'])
        return f"Started on {start_time.strftime('%d %b %Y %H:%M %Z')} by {details['submittedBy']}"

    html = f"<p>Experiment: {format_experiment(parent_run.experiment)}"
    html += f"\n<br>Parent run: {format_run(parent_run)}"

    html += "\n<ul>"
    for k, child_run in enumerate(sorted(parent_run.get_children(), key=lambda r: r.id)):
        html += f"\n<li>Child {k}: {format_run(child_run)}</li>"
    html += "\n</ul>"

    html += f"\n<p>{format_submission_info(parent_run)}"
    html += f"\n<p>Command-line arguments: <code>{parent_run.get_tags()['commandline_args']}</code>"
    return html


def get_child_runs_hyperparams(metrics_df: pd.DataFrame) -> Dict[int, Dict]:
    """
    Get the hyperparameters of each child run from the metrics dataframe.
    :param: metrics_df: Metrics dataframe, as returned by :py:func:`collect_hyperdrive_metrics()` and
        :py:func:`~health_azure.aggregate_hyperdrive_metrics()`.
    :return: A dictionary of hyperparameter dictionaries for the child runs.
    """
    hyperparams_children = {}
    for child_index in metrics_df.columns:
        hyperparams = metrics_df[child_index][AMLMetricsJsonKey.HYPERPARAMS]
        hyperparams_dict = dict(zip(hyperparams[AMLMetricsJsonKey.NAME], hyperparams[AMLMetricsJsonKey.VALUE]))
        hyperparams_children[child_index] = hyperparams_dict
    return hyperparams_children


def collect_class_info(hyperparams_children: Dict[int, Dict]) -> Tuple[int, List[str]]:
    """
    Get the class names from the hyperparameters of child runs.
    :param hyperparams_children: Dict of hyperparameter dicts, as returned by :py:func:`get_child_runs_hyperparams()`.
    :return: Number of classes and list of class names.
    """
    hyperparams_single_run = list(hyperparams_children.values())[0]
    num_classes = int(hyperparams_single_run[AMLMetricsJsonKey.N_CLASSES])
    class_names = hyperparams_single_run[AMLMetricsJsonKey.CLASS_NAMES]
    if class_names == "None":
        class_names = None
    else:
        # Remove [,], and quotation marks from the string of class names
        class_names = [name.lstrip() for name in class_names[1:-1].replace("'", "").split(',')]
    class_names = validate_class_names(class_names=class_names, n_classes=num_classes)
    return (num_classes, list(class_names))


def get_max_epochs(hyperparams_children: Dict[int, Dict]) -> Dict[int, int]:
    """
    Get the maximum number of epochs for each round from the metrics dataframe.
    :param hyperparams_children: Dict of hyperparameter dicts, as returned by :py:func:`get_child_runs_hyperparams()`.
    :return: Dictionary with the number of epochs in each hyperdrive run.
    """
    max_epochs_dict = {}
    for child_index in hyperparams_children.keys():
        max_epochs_dict[child_index] = int(hyperparams_children[child_index][AMLMetricsJsonKey.MAX_EPOCHS])
    return max_epochs_dict



#  -------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  -------------------------------------------------------------------------------------------
"""
This script can be used to create a montage of slides, given a slides dataset or a folder with images.

For a full documentation of the parameters, run `python create_montage.py --help`
"""
import logging
from pathlib import Path
import sys
from typing import Optional


current_file = Path(__file__).absolute()
repository_root = current_file.parent.parent.parent.parent.parent
folders_to_add = [
    repository_root / "hi-ml" / "src",
    repository_root / "hi-ml-azure" / "src",
    repository_root / "hi-ml-cpath" / "src",
]
for folder in folders_to_add:
    assert folder.is_dir()
    sys.path.insert(0, str(folder))

from health_azure.himl import submit_to_azure_if_needed, DatasetConfig  # noqa
from health_azure.logging import logging_to_stdout  # noqa
from health_cpath.utils.montage import create_config_from_args  # noqa


def main() -> None:
    config = create_config_from_args()
    logging_to_stdout()
    submit_to_azureml = config.cluster != ""
    if config.dataset.strip() == "":
        raise ValueError("Please provide a dataset name via --dataset")
    elif config.dataset.startswith("/"):
        if submit_to_azureml:
            raise ValueError("Cannot submit to AzureML if dataset is a local folder")
        input_folder: Optional[Path] = Path(config.dataset)
    else:
        logging.info(f"In AzureML use mounted dataset '{config.dataset}' in datastore {config.datastore}")
        input_dataset = DatasetConfig(name=config.dataset, datastore=config.datastore, use_mounting=True)
        logging.info(f"Submitting to AzureML, running on cluster {config.cluster}")
        run_info = submit_to_azure_if_needed(
            entry_script=current_file,
            snapshot_root_directory=repository_root,
            compute_cluster_name=config.cluster,
            conda_environment_file=config.conda_env,
            submit_to_azureml=submit_to_azureml,
            input_datasets=[input_dataset],
            strictly_aml_v1=True,
            docker_shm_size=config.docker_shm_size,
            wait_for_completion=config.wait_for_completion,
            workspace_config_file=config.workspace_config_path,
            display_name=config.display_name,
        )
        input_folder = run_info.input_datasets[0]

    assert input_folder is not None
    config.create_montage(input_folder)


if __name__ == "__main__":
    main()



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

from pathlib import Path
from typing import Any, Callable, Optional, Tuple, Union

import pandas as pd

from torchvision.datasets.vision import VisionDataset

from health_cpath.datasets.base_dataset import TilesDataset
from health_cpath.models.transforms import load_pil_image
from health_cpath.datasets.dataset_return_index import DatasetWithReturnIndex


class TcgaCrck_TilesDataset(TilesDataset):
    """Dataset class for loading TCGA-CRCk tiles.

    Iterating over this dataset returns a dictionary containing:
    - `'slide_id'` (str): parent slide ID
    - `'tile_id'` (str)
    - `'image'` (`PIL.Image`): RGB tile
    - `'label'` (str): MSS (0) vs MSIMUT (1)
    """

    TILE_X_COLUMN = TILE_Y_COLUMN = None  # no tile coordinates available
    # This dataset conforms to all other defaults in TilesDataset


class TcgaCrck_TilesDatasetReturnImageLabel(VisionDataset):
    """
    Any dataset used in SSL needs to return a tuple where the first element is the image and the second is a
    class label.
    """

    def __init__(
        self,
        root: Union[str, Path],
        dataset_csv: Optional[Union[str, Path]] = None,
        dataset_df: Optional[pd.DataFrame] = None,
        train: Optional[bool] = None,
        transform: Optional[Callable] = None,
        **kwargs: Any
    ) -> None:
        super().__init__(root=root, transform=transform)
        self.base_dataset = TcgaCrck_TilesDataset(
            root=root, dataset_csv=dataset_csv, dataset_df=dataset_df, train=train
        )

    def __getitem__(self, index: int) -> Tuple:  # type: ignore
        sample = self.base_dataset[index]
        # TODO change to a meaningful evaluation
        image = load_pil_image(sample[self.base_dataset.IMAGE_COLUMN])
        if self.transform:
            image = self.transform(image)
        return image, sample[self.base_dataset.label_column]

    def __len__(self) -> int:
        return len(self.base_dataset)


class TcgaCrck_TilesDatasetWithReturnIndex(DatasetWithReturnIndex, TcgaCrck_TilesDatasetReturnImageLabel):
    """
    Any dataset used in SSL needs to inherit from DataClassBaseWithReturnIndex as well as VisionData.
    This class is just a shorthand notation for this double inheritance. Please note that this class needs
    to override __getitem__(), this is why we need a separate PandaTilesDatasetReturnImageLabel.
    """

    @property
    def num_classes(self) -> int:
        return 2



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------
from typing import Any, Optional, Set
from health_azure.utils import is_running_in_azure_ml, create_from_matching_params
from health_cpath.configs.classification.BaseMIL import BaseMIL, BaseMILSlides, BaseMILTiles
from health_cpath.configs.run_ids import innereye_ssl_checkpoint_binary
from health_cpath.datamodules.panda_module import PandaSlidesDataModule, PandaTilesDataModule
from health_cpath.datasets.default_paths import PANDA_5X_TILES_DATASET_ID, PANDA_DATASET_ID
from health_cpath.datasets.panda_dataset import PandaDataset
from health_cpath.datasets.panda_tiles_dataset import PandaTilesDataset
from health_cpath.models.encoders import HistoSSLEncoder, ImageNetSimCLREncoder, Resnet18, SSLEncoder
from health_cpath.preprocessing.loading import LoadingParams, ROIType, WSIBackend
from health_cpath.utils.naming import PlotOption, SlideKey
from health_cpath.utils.wsi_utils import TilingParams
from health_ml.networks.layers.attention_layers import AttentionLayer
from health_ml.utils.checkpoint_utils import CheckpointParser


class BaseDeepSMILEPanda(BaseMIL):
    """Base class for DeepSMILEPanda common configs between tiles and slides piplines."""

    def __init__(self, **kwargs: Any) -> None:
        default_kwargs = dict(
            # declared in BaseMIL:
            pool_type=AttentionLayer.__name__,
            num_transformer_pool_layers=4,
            num_transformer_pool_heads=4,
            tune_encoder=False,
            # average number of tiles is 56 for PANDA
            encoding_chunk_size=60,
            max_bag_size=56,
            max_bag_size_inf=0,
            # declared in TrainerParams:
            max_epochs=200,
            # declared in OptimizerParams:
            l_rate=5e-4,
            weight_decay=1e-4,
            adam_betas=(0.9, 0.99),
            # loading params:
            backend=WSIBackend.CUCIM,
            roi_type=ROIType.WHOLE,
            image_key=SlideKey.IMAGE,
            mask_key=SlideKey.MASK,
        )
        default_kwargs.update(kwargs)
        super().__init__(**default_kwargs)
        self.class_names = ["ISUP 0", "ISUP 1", "ISUP 2", "ISUP 3", "ISUP 4", "ISUP 5"]
        if not is_running_in_azure_ml():
            self.max_epochs = 2

    def get_test_plot_options(self) -> Set[PlotOption]:
        plot_options = super().get_test_plot_options()
        plot_options.update([PlotOption.SLIDE_THUMBNAIL, PlotOption.ATTENTION_HEATMAP])
        return plot_options


class DeepSMILETilesPanda(BaseMILTiles, BaseDeepSMILEPanda):
    """DeepSMILETilesPanda is derived from BaseMILTiles and BaseDeepSMILEPanda to inherit common behaviors from both
    tiles basemil and panda specific configuration.

    `tune_encoder` sets the fine-tuning mode of the encoder. For fine-tuning the encoder, batch_size = 2
    runs on multiple GPUs with ~ 6:24 min/epoch (train) and ~ 00:50 min/epoch (validation).
    """

    def __init__(self, **kwargs: Any) -> None:
        default_kwargs = dict(
            # declared in BaseMILTiles:
            is_caching=False,
            batch_size=8,
            batch_size_inf=8,
            azure_datasets=[PANDA_5X_TILES_DATASET_ID, PANDA_DATASET_ID],
        )
        default_kwargs.update(kwargs)
        super().__init__(**default_kwargs)

    def setup(self) -> None:
        BaseMILTiles.setup(self)

    def get_data_module(self) -> PandaTilesDataModule:
        return PandaTilesDataModule(
            root_path=self.local_datasets[0],
            batch_size=self.batch_size,
            batch_size_inf=self.batch_size_inf,
            max_bag_size=self.max_bag_size,
            max_bag_size_inf=self.max_bag_size_inf,
            transforms_dict=self.get_transforms_dict(PandaTilesDataset.IMAGE_COLUMN),
            cache_mode=self.cache_mode,
            precache_location=self.precache_location,
            cache_dir=self.cache_dir,
            crossval_count=self.crossval_count,
            crossval_index=self.crossval_index,
            dataloader_kwargs=self.get_dataloader_kwargs(),
            seed=self.get_effective_random_seed(),
            pl_replace_sampler_ddp=self.pl_replace_sampler_ddp,
        )

    def get_slides_dataset(self) -> Optional[PandaDataset]:
        return PandaDataset(root=self.local_datasets[1])  # type: ignore


class TilesPandaImageNetMIL(DeepSMILETilesPanda):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(encoder_type=Resnet18.__name__, **kwargs)


class TilesPandaImageNetSimCLRMIL(DeepSMILETilesPanda):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(encoder_type=ImageNetSimCLREncoder.__name__, **kwargs)


class TilesPandaSSLMIL(DeepSMILETilesPanda):
    def __init__(self, **kwargs: Any) -> None:
        # If no SSL checkpoint is provided, use the default one
        self.ssl_checkpoint = self.ssl_checkpoint or CheckpointParser(innereye_ssl_checkpoint_binary)
        super().__init__(encoder_type=SSLEncoder.__name__, **kwargs)


class TilesPandaHistoSSLMIL(DeepSMILETilesPanda):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(encoder_type=HistoSSLEncoder.__name__, **kwargs)


class DeepSMILESlidesPanda(BaseMILSlides, BaseDeepSMILEPanda):
    """DeepSMILESlidesPanda is derived from BaseMILSlides and BaseDeeppSMILEPanda to inherits common behaviors from both
    slides basemil and panda specific configuration.
    """

    def __init__(self, **kwargs: Any) -> None:
        default_kwargs = dict(
            level=1,
            tile_size=224,
            background_val=255,
            azure_datasets=[PANDA_DATASET_ID],
        )
        default_kwargs.update(kwargs)
        super().__init__(**default_kwargs)

    def setup(self) -> None:
        BaseMILSlides.setup(self)

    def get_dataloader_kwargs(self) -> dict:
        return dict(multiprocessing_context="spawn", **super().get_dataloader_kwargs())

    def get_data_module(self) -> PandaSlidesDataModule:
        return PandaSlidesDataModule(
            root_path=self.local_datasets[0],
            batch_size=self.batch_size,
            batch_size_inf=self.batch_size_inf,
            max_bag_size=self.max_bag_size,
            max_bag_size_inf=self.max_bag_size_inf,
            tiling_params=create_from_matching_params(self, TilingParams),
            loading_params=create_from_matching_params(self, LoadingParams),
            seed=self.get_effective_random_seed(),
            transforms_dict=self.get_transforms_dict(SlideKey.IMAGE),
            crossval_count=self.crossval_count,
            crossval_index=self.crossval_index,
            dataloader_kwargs=self.get_dataloader_kwargs(),
            pl_replace_sampler_ddp=self.pl_replace_sampler_ddp,
        )

    def get_slides_dataset(self) -> PandaDataset:
        return PandaDataset(root=self.local_datasets[0])  # type: ignore


class SlidesPandaImageNetMIL(DeepSMILESlidesPanda):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(encoder_type=Resnet18.__name__, **kwargs)


class SlidesPandaImageNetSimCLRMIL(DeepSMILESlidesPanda):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(encoder_type=ImageNetSimCLREncoder.__name__, **kwargs)


class SlidesPandaSSLMIL(DeepSMILESlidesPanda):
    def __init__(self, **kwargs: Any) -> None:
        # If no SSL checkpoint is provided, use the default one
        self.ssl_checkpoint = self.ssl_checkpoint or CheckpointParser(innereye_ssl_checkpoint_binary)
        super().__init__(encoder_type=SSLEncoder.__name__, **kwargs)


class SlidesPandaHistoSSLMIL(DeepSMILESlidesPanda):
    def __init__(self, **kwargs: Any) -> None:
        super().__init__(encoder_type=HistoSSLEncoder.__name__, **kwargs)



#  -------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  -------------------------------------------------------------------------------------------

from pathlib import Path
from typing import Dict, List, Optional

import pandas as pd
from matplotlib import pyplot as plt

from health_azure.utils import get_aml_run_from_run_id, get_workspace
from health_ml.utils.reports import HTMLReport
from health_cpath.utils.analysis_plot_utils import (
    add_training_curves_legend,
    plot_confusion_matrices,
    plot_hyperdrive_roc_and_pr_curves,
    plot_hyperdrive_training_curves,
)
from health_cpath.utils.output_utils import AML_LEGACY_TEST_OUTPUTS_CSV, AML_TEST_OUTPUTS_CSV, AML_VAL_OUTPUTS_CSV
from health_cpath.utils.report_utils import (
    collect_hyperdrive_metrics,
    collect_hyperdrive_outputs,
    child_runs_have_val_and_test_outputs,
    get_best_epoch_metrics,
    get_best_epochs,
    get_child_runs_hyperparams,
    get_hyperdrive_metrics_table,
    get_formatted_run_info,
    collect_class_info,
    get_max_epochs,
    download_hyperdrive_metrics_if_required,
)
from health_cpath.utils.naming import MetricsKey, ModelKey


def generate_html_report(
    parent_run_id: str,
    output_dir: Path,
    workspace_config_path: Optional[Path] = None,
    include_test: bool = False,
    overwrite: bool = False,
    hyperdrive_arg_name: str = "crossval_index",
    primary_metric: str = MetricsKey.AUROC,
) -> None:
    """
    Function to generate an HTML report of a Hyperdrive AML run (e.g., cross validation, different random seeds, ...).

    :param run_id: The parent Hyperdrive run ID.
    :param output_dir: Directory where to download Azure ML data and save the report.
    :param workspace_config_path: Path to Azure ML workspace config.json file.
        If omitted, will try to load default workspace.
    :param include_test: Include test results in the generated report.
    :param overwrite: Forces (re)download of metrics and output files, even if they already exist locally.
    :param hyperdrive_arg_name: Name of the Hyperdrive argument used for indexing the child runs.
        Default `crossval_index`.
    :param primary_metric: Name of the reference metric to optimise. Default `MetricsKey.AUROC`.
    """
    aml_workspace = get_workspace(workspace_config_path=workspace_config_path)
    parent_run = get_aml_run_from_run_id(parent_run_id, aml_workspace=aml_workspace)
    report_dir = output_dir / parent_run.display_name
    report_dir.mkdir(parents=True, exist_ok=True)

    report = HTMLReport(output_folder=report_dir)

    report.add_text(get_formatted_run_info(parent_run))

    report.add_heading("Azure ML metrics", level=2)

    # Download metrics from AML. Can take several seconds for each child run
    metrics_json = download_hyperdrive_metrics_if_required(
        parent_run_id, report_dir, aml_workspace, overwrite=overwrite, hyperdrive_arg_name=hyperdrive_arg_name
    )

    # Get metrics dataframe from the downloaded json file
    metrics_df = collect_hyperdrive_metrics(metrics_json=metrics_json)

    hyperparameters_children = get_child_runs_hyperparams(metrics_df)
    max_epochs_dict = get_max_epochs(hyperparams_children=hyperparameters_children)
    best_epochs = get_best_epochs(
        metrics_df=metrics_df,
        primary_metric=f'{ModelKey.VAL}/{primary_metric}',
        max_epochs_dict=max_epochs_dict,
        maximise=True,
    )

    # Add training curves for loss and AUROC (train and val.)
    render_training_curves(
        report,
        heading="Training curves",
        level=3,
        metrics_df=metrics_df,
        best_epochs=best_epochs,
        report_dir=report_dir,
        primary_metric=primary_metric,
    )

    # Get metrics list with class names
    num_classes, class_names = collect_class_info(hyperparams_children=hyperparameters_children)

    base_metrics_list: List[str] = [
        MetricsKey.ACC,
        MetricsKey.AUROC,
        MetricsKey.AVERAGE_PRECISION,
        MetricsKey.COHENKAPPA,
    ]
    if num_classes > 1:
        base_metrics_list += [MetricsKey.ACC_MACRO, MetricsKey.ACC_WEIGHTED]
    else:
        base_metrics_list += [MetricsKey.F1, MetricsKey.PRECISION, MetricsKey.RECALL, MetricsKey.SPECIFICITY]

    base_metrics_list += class_names

    # Add tables with relevant metrics (val. and test)
    render_metrics_table(
        report,
        heading=f"Validation metrics (best epoch based on maximum validation {primary_metric})",
        level=3,
        metrics_df=metrics_df,
        best_epochs=best_epochs,
        base_metrics_list=base_metrics_list,
        metrics_prefix=f'{ModelKey.VAL}/',
    )

    if include_test:
        render_metrics_table(
            report,
            heading="Test metrics",
            level=3,
            metrics_df=metrics_df,
            best_epochs=None,
            base_metrics_list=base_metrics_list,
            metrics_prefix=f'{ModelKey.TEST}/',
        )

    # Get output data frames if available
    try:
        has_val_and_test_outputs = child_runs_have_val_and_test_outputs(parent_run)
        if has_val_and_test_outputs:
            output_filename_val = AML_VAL_OUTPUTS_CSV
            outputs_dfs_val = collect_hyperdrive_outputs(
                parent_run_id=parent_run_id,
                download_dir=report_dir,
                aml_workspace=aml_workspace,
                output_filename=output_filename_val,
                overwrite=overwrite,
                hyperdrive_arg_name=hyperdrive_arg_name,
            )
            if include_test:
                output_filename_test = AML_TEST_OUTPUTS_CSV if has_val_and_test_outputs else AML_LEGACY_TEST_OUTPUTS_CSV
                outputs_dfs_test = collect_hyperdrive_outputs(
                    parent_run_id=parent_run_id,
                    download_dir=report_dir,
                    aml_workspace=aml_workspace,
                    output_filename=output_filename_test,
                    overwrite=overwrite,
                    hyperdrive_arg_name=hyperdrive_arg_name,
                )

        if num_classes == 1:
            # Currently ROC and PR curves rendered only for binary case
            # TODO: Enable rendering of multi-class ROC and PR curves
            report.add_heading("ROC and PR curves", level=2)
            if has_val_and_test_outputs:
                # Add val. ROC and PR curves
                render_roc_and_pr_curves(
                    report=report,
                    heading="Validation ROC and PR curves",
                    level=3,
                    report_dir=report_dir,
                    outputs_dfs=outputs_dfs_val,
                    prefix=f'{ModelKey.VAL}_',
                )
            if include_test:
                # Add test ROC and PR curves
                render_roc_and_pr_curves(
                    report=report,
                    heading="Test ROC and PR curves",
                    level=3,
                    report_dir=report_dir,
                    outputs_dfs=outputs_dfs_test,
                    prefix=f'{ModelKey.TEST}_',
                )

        # Add confusion matrices for each fold
        report.add_heading("Confusion matrices", level=2)
        if has_val_and_test_outputs:
            # Add val. confusion matrices
            render_confusion_matrices(
                report=report,
                heading="Validation confusion matrices",
                level=3,
                class_names=class_names,
                report_dir=report_dir,
                outputs_dfs=outputs_dfs_val,
                prefix=f'{ModelKey.VAL}_',
            )

        if include_test:
            # Add test confusion matrices
            render_confusion_matrices(
                report=report,
                heading="Test confusion matrices",
                level=3,
                class_names=class_names,
                report_dir=report_dir,
                outputs_dfs=outputs_dfs_test,
                prefix=f'{ModelKey.TEST}_',
            )

    except ValueError as e:
        print(e)
        print("Since all expected output files were not found, skipping ROC-PR curves and confusion matrices.")

    # TODO: Add qualitative model outputs
    # report.add_heading("Qualitative model outputs", level=2)

    print(f"Rendering report to: {report.report_path_html.resolve()}")
    report.render()


def render_training_curves(
    report: HTMLReport,
    heading: str,
    level: int,
    metrics_df: pd.DataFrame,
    best_epochs: Optional[Dict[int, int]],
    report_dir: Path,
    primary_metric: str = MetricsKey.AUROC,
) -> None:
    """
    Function to render training curves for HTML reports.

    :param report: HTML report to perform the rendering.
    :param heading: Heading of the section.
    :param level: Level of HTML heading (e.g. sub-section, sub-sub-section) corresponding to HTML heading levels.
    :param metrics_df: Metrics dataframe, as returned by :py:func:`collect_hyperdrive_metrics()` and
        :py:func:`~health_azure.aggregate_hyperdrive_metrics()`.
    :param best_epochs: Dictionary mapping each hyperdrive child index to its best epoch.
    :param report_dir: Directory of the HTML report.
    :param primary_metric: Primary metric name. Default is AUROC.
    """
    report.add_heading(heading, level=level)
    metrics = {"loss_epoch", MetricsKey.AUROC.value, primary_metric}
    fig, axs = plt.subplots(1, len(metrics), figsize=(5 * len(metrics), 4))
    for i, metric in enumerate(metrics):
        plot_hyperdrive_training_curves(
            metrics_df,
            train_metric=f'{ModelKey.TRAIN}/{metric}',
            val_metric=f'{ModelKey.VAL}/{metric}',
            ylabel=metric,
            best_epochs=best_epochs,
            ax=axs[i],
        )
    add_training_curves_legend(fig, include_best_epoch=True)
    training_curves_fig_path = report_dir / "training_curves.png"
    fig.savefig(training_curves_fig_path, bbox_inches='tight')
    report.add_images([training_curves_fig_path], base64_encode=True)


def render_metrics_table(
    report: HTMLReport,
    heading: str,
    level: int,
    metrics_df: pd.DataFrame,
    best_epochs: Optional[Dict[int, int]],
    base_metrics_list: List[str],
    metrics_prefix: str,
) -> None:
    """
    Function to render metrics table for HTML reports.

    :param report: HTML report to perform the rendering.
    :param heading: Heading of the section.
    :param level: Level of HTML heading (e.g. sub-section, sub-sub-section) corresponding to HTML heading levels.
    :param metrics_df: Metrics dataframe, as returned by :py:func:`collect_hyperdrive_metrics()` and
        :py:func:`~health_azure.aggregate_hyperdrive_metrics()`.
    :param base_metrics_list: List of metric names to include in the table.
    :param best_epochs: Dictionary mapping each hyperdrive child index to its best epoch.
    :param metrics_prefix: Prefix to add to the metrics names (e.g. `val`, `test`)
    """
    report.add_heading(heading, level=level)
    metrics_list = [metrics_prefix + metric for metric in base_metrics_list]
    if best_epochs:
        metrics_df = get_best_epoch_metrics(metrics_df, metrics_list, best_epochs)
    metrics_table = get_hyperdrive_metrics_table(metrics_df, metrics_list)
    report.add_tables([metrics_table])


def render_roc_and_pr_curves(
    report: HTMLReport,
    heading: str,
    level: int,
    report_dir: Path,
    outputs_dfs: Dict[int, pd.DataFrame],
    prefix: str = '',
) -> None:
    """
    Function to render ROC and PR curves for HTML reports.

    :param report: HTML report to perform the rendering.
    :param heading: Heading of the section.
    :param level: Level of HTML heading (e.g. sub-section, sub-sub-section) corresponding to HTML heading levels.
    :param report_dir: Local directory where the report is stored.
    :param outputs_dfs: A dictionary of dataframes with the sorted hyperdrive child runs indices as keys.
    :param prefix: Prefix to add to the figures saved (e.g. `val`, `test`).
    """
    report.add_heading(heading, level=level)
    fig = plot_hyperdrive_roc_and_pr_curves(outputs_dfs, scores_column='prob_class1')
    roc_pr_curves_fig_path = report_dir / f"{prefix}roc_pr_curves.png"
    fig.savefig(roc_pr_curves_fig_path, bbox_inches='tight')
    report.add_images([roc_pr_curves_fig_path], base64_encode=True)


def render_confusion_matrices(
    report: HTMLReport,
    heading: str,
    level: int,
    class_names: List[str],
    report_dir: Path,
    outputs_dfs: Dict[int, pd.DataFrame],
    prefix: str = '',
) -> None:
    """
    Function to render confusion matrices for HTML reports.

    :param report: HTML report to perform the rendering.
    :param heading: Heading of the section.
    :param level: Level of HTML heading (e.g. sub-section, sub-sub-section) corresponding to HTML heading levels.
    :param class_names: Names of classes.
    :param report_dir: Local directory where the report is stored.
    :param outputs_dfs: A dictionary of dataframes with the sorted hyperdrive child runs indices as keys.
    :param prefix: Prefix to add to the figures saved (e.g. `val`, `test`).
    """
    report.add_heading(heading, level=level)
    fig = plot_confusion_matrices(hyperdrive_dfs=outputs_dfs, class_names=class_names)
    confusion_matrices_fig_path = report_dir / f"{prefix}confusion_matrices.png"
    fig.savefig(confusion_matrices_fig_path, bbox_inches='tight')
    report.add_images([confusion_matrices_fig_path], base64_encode=True)


if __name__ == "__main__":
    """
    Usage example from CLI:
    python generate_hyperdrive_report.py \
    --run_id <insert AML run ID here> \
    --output_dir outputs \
    --include_test
    """

    from argparse import ArgumentParser

    parser = ArgumentParser()
    parser.add_argument('--run_id', help="The parent Hyperdrive run ID.")
    parser.add_argument('--output_dir', help="Directory where to download Azure ML data and save the report.")
    parser.add_argument(
        '--workspace_config',
        help="Path to Azure ML workspace config.json file. If omitted, will try to load default workspace.",
    )
    parser.add_argument(
        '--include_test', action='store_true', help="Opt-in flag to include test results in the generated report."
    )
    parser.add_argument(
        '--overwrite',
        action='store_true',
        help="Forces (re)download of metrics and output files, even if they already exist locally.",
    )
    parser.add_argument(
        "--hyper_arg_name",
        default="crossval_index",
        help="Name of the Hyperdrive argument used for indexing the child runs.",
    )
    parser.add_argument("--primary_metric", default=MetricsKey.AUROC, help="Name of the reference metric to optimise.")
    args = parser.parse_args()

    if args.output_dir is None:
        args.output_dir = Path.cwd() / "outputs"
    workspace_config = Path(args.workspace_config).resolve() if args.workspace_config else None

    print(f"Output dir: {Path(args.output_dir).resolve()}")
    if workspace_config is not None:
        if not workspace_config.is_file():
            raise ValueError(f"Specified workspace config file does not exist: {workspace_config}")
        print(f"Workspace config: {workspace_config}")

    generate_html_report(
        parent_run_id=args.run_id,
        output_dir=Path(args.output_dir),
        workspace_config_path=workspace_config,
        include_test=args.include_test,
        overwrite=args.overwrite,
        hyperdrive_arg_name=args.hyper_arg_name,
        primary_metric=args.primary_metric,
    )



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

"""
Configuration of data modules for PANDA DeepMIL experiments.
Ref: Myronenko et al. 2021 (https://link.springer.com/chapter/10.1007/978-3-030-87237-3_32)
"""

from typing import Tuple

from health_ml.utils.split_dataset import DatasetSplits

from health_cpath.datamodules.base_module import SlidesDataModule
from health_cpath.datasets.panda_dataset import PandaDataset


class PandaSlidesDataModuleBenchmark(SlidesDataModule):
    """PandaSlidesDataModuleBenchmark is the child class of SlidesDataModule specific to PANDA dataset
    Method get_splits() returns the train, val, test splits from the PANDA dataset
    """

    def get_splits(self) -> Tuple[PandaDataset, PandaDataset, PandaDataset]:
        dataset = PandaDataset(self.root_path)
        splits = DatasetSplits.from_proportions(
            dataset.dataset_df.reset_index(),
            proportion_train=0.8,
            proportion_test=0.0,
            proportion_val=0.2,
            subject_column=dataset.slide_id_column,
        )

        if self.crossval_count > 1:
            # Function get_k_fold_cross_validation_splits() will concatenate train and val splits
            splits = splits.get_k_fold_cross_validation_splits(self.crossval_count)[self.crossval_index]

        return (
            PandaDataset(self.root_path, dataset_df=splits.train),
            PandaDataset(self.root_path, dataset_df=splits.val),
            PandaDataset(self.root_path, dataset_df=splits.val),
        )



#  ------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  ------------------------------------------------------------------------------------------

import os
import param
import logging
import warnings
from typing import Any, Callable, Dict, List, Optional, Sequence, Set, Union

from pathlib import Path
from monai.transforms import Compose
from pytorch_lightning.callbacks import Callback
from pytorch_lightning.callbacks.model_checkpoint import ModelCheckpoint
import torch

from health_azure.utils import create_from_matching_params
from health_cpath.preprocessing.loading import LoadingParams
from health_cpath.utils.callbacks import LossAnalysisCallback, LossCallbackParams
from health_cpath.utils.wsi_utils import TilingParams

from health_ml.deep_learning_config import OptimizerParams
from health_ml.experiment_config import RunnerMode
from health_ml.lightning_container import LightningContainer
from health_ml.utils.checkpoint_utils import CheckpointParser

from health_cpath.datamodules.base_module import CacheLocation, CacheMode, HistoDataModule
from health_cpath.datasets.base_dataset import SlidesDataset
from health_cpath.models.deepmil import DeepMILModule
from health_cpath.models.transforms import EncodeTilesBatchd, LoadTilesBatchd
from health_cpath.utils.deepmil_utils import ClassifierParams, EncoderParams, PoolingParams
from health_cpath.utils.output_utils import DeepMILOutputsHandler
from health_cpath.utils.naming import MetricsKey, PlotOption, SlideKey, ModelKey
from health_cpath.utils.tiles_selection_utils import TilesSelector


class BaseMIL(LightningContainer, LoadingParams, EncoderParams, PoolingParams, ClassifierParams, LossCallbackParams):
    """BaseMIL is an abstract container defining basic functionality for running MIL experiments in both slides and
    tiles settings. It is responsible for instantiating the encoder and pooling layer. Subclasses should define the
    full DeepMIL model depending on the type of dataset (tiles/slides based).
    """

    class_names: Optional[Sequence[str]] = param.List(
        None, item_type=str, doc="List of class names. If `None`, defaults to `('0', '1', ...)`."
    )
    # Data module parameters:
    batch_size: int = param.Integer(16, bounds=(1, None), doc="Number of slides to load per batch.")
    batch_size_inf: int = param.Integer(16, bounds=(1, None), doc="Number of slides per batch during inference.")
    max_bag_size: int = param.Integer(
        1000,
        bounds=(0, None),
        doc="Upper bound on number of tiles in each loaded bag during training stage. "
        "If 0 (default), will return all samples in each bag. "
        "If > 0, bags larger than `max_bag_size` will yield "
        "random subsets of instances.",
    )
    max_bag_size_inf: int = param.Integer(
        0,
        bounds=(0, None),
        doc="Upper bound on number of tiles in each loaded bag during "
        "validation and test stages."
        "If 0 (default), will return all samples in each bag. "
        "If > 0 , bags larger than `max_bag_size_inf` will yield "
        "random subsets of instances.",
    )
    # Outputs Handler parameters:
    num_top_slides: int = param.Integer(
        10,
        bounds=(0, None),
        doc="Number of slides to select when saving top and "
        "bottom tiles. If set to 10 (default), it selects 10 "
        "top and 10 bottom slides. To disable tiles plots "
        "saving, set `num_top_slides=0`",
    )
    num_top_tiles: int = param.Integer(
        12,
        bounds=(1, None),
        doc="Number of tiles to select when saving top and bottom"
        "tiles. If set to 12 (default), it saves 12 top and 12"
        "bottom tiles.",
    )
    primary_val_metric: MetricsKey = param.ClassSelector(
        default=MetricsKey.AUROC,
        class_=MetricsKey,
        doc="Primary validation metric to track for checkpointing and generating outputs.",
    )
    maximise_primary_metric: bool = param.Boolean(
        True, doc="Whether the primary validation metric should be maximised (otherwise minimised)."
    )
    max_num_workers: int = param.Integer(
        10,
        bounds=(0, None),
        doc="The maximum number of worker processes for dataloaders. Dataloaders use"
        "a heuristic num_cpus/num_gpus to set the number of workers, which can be"
        "very high for small num_gpus. This parameters sets an upper bound.",
    )
    save_intermediate_outputs: bool = param.Boolean(
        True, doc="Whether to save intermediate validation outputs during training."
    )
    stratify_plots_by: Optional[str] = param.String(
        None, doc="Name of metadata field to stratify output plots (PR curve, ROC curve)."
    )

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.data_module: Optional[HistoDataModule] = None
        self.run_extra_val_epoch = True  # Enable running an additional validation step to save tiles/slides thumbnails
        metric_optim = "max" if self.maximise_primary_metric else "min"
        self.best_checkpoint_filename = f"checkpoint_{metric_optim}_val_{self.primary_val_metric.value}"
        self.best_checkpoint_filename_with_suffix = self.best_checkpoint_filename + ".ckpt"
        self.validate()
        if not self.pl_replace_sampler_ddp and self.max_num_gpus > 1:
            logging.info(
                "Replacing sampler with `DistributedSampler` is disabled. Make sure to set your own DDP sampler"
            )
        self.ignore_pl_warnings()

    def validate(self) -> None:
        super().validate()
        EncoderParams.validate(self)
        if not any([self.tune_encoder, self.tune_pooling, self.tune_classifier]) and not self.run_inference_only:
            raise ValueError(
                "At least one of the encoder, pooling or classifier should be fine tuned. Turn on one of the tune "
                "arguments `tune_encoder`, `tune_pooling`, `tune_classifier`. Otherwise, activate inference only "
                "mode via `run_inference_only` flag."
            )
        if (
            any([self.pretrained_encoder, self.pretrained_pooling, self.pretrained_classifier])
            and not self.src_checkpoint
        ):
            raise ValueError(
                "You need to specify a source checkpoint, to use a pretrained encoder, pooling or classifier."
                f" {CheckpointParser.INFO_MESSAGE}"
            )
        if (
            self.tune_encoder
            and self.encoding_chunk_size < self.max_bag_size
            and self.pl_sync_batchnorm
            and self.max_num_gpus > 1
        ):
            raise ValueError(
                "The encoding chunk size should be at least as large as the maximum bag size when fine tuning the "
                "encoder. You might encounter Batch Norm synchronization issues if the chunk size is smaller than "
                "the maximum bag size causing the processes to hang silently. This is due to the encoder being called "
                "different number of times on each device, which cause Batch Norm running statistics to be updated "
                "inconsistently across processes. In case you can't increase the `encoding_chunk_size` any further, set"
                " `pl_sync_batchnorm=False` to simply skip Batch Norm synchronization across devices. Note that this "
                "might come with some performance penalty."
            )

    @property
    def cache_dir(self) -> Path:
        return Path(f"/tmp/himl_cache/{self.__class__.__name__}-{self.encoder_type}/")

    def get_test_plot_options(self) -> Set[PlotOption]:
        options = {PlotOption.HISTOGRAM, PlotOption.CONFUSION_MATRIX}
        if self.num_top_slides > 0:
            options.update([PlotOption.TOP_BOTTOM_TILES, PlotOption.ATTENTION_HISTOGRAM])
        return options

    def get_val_plot_options(self) -> Set[PlotOption]:
        """Override this method if you want to produce validation plots at each epoch. By default, at the end of the
        training an extra validation epoch is run where val_plot_options = test_plot_options
        """
        return set()

    def get_outputs_handler(self) -> DeepMILOutputsHandler:
        n_classes = self.get_num_classes()
        outputs_handler = DeepMILOutputsHandler(
            outputs_root=self.outputs_folder,
            n_classes=n_classes,
            tile_size=self.tile_size,
            class_names=self.class_names,
            primary_val_metric=self.primary_val_metric,
            maximise=self.maximise_primary_metric,
            val_plot_options=self.get_val_plot_options(),
            test_plot_options=self.get_test_plot_options(),
            loading_params=create_from_matching_params(self, LoadingParams),
            save_intermediate_outputs=self.save_intermediate_outputs,
            stratify_plots_by=self.stratify_plots_by,
        )
        if self.num_top_slides > 0:
            outputs_handler.tiles_selector = TilesSelector(
                n_classes=n_classes, num_slides=self.num_top_slides, num_tiles=self.num_top_tiles
            )
        return outputs_handler

    def get_callbacks(self) -> List[Callback]:
        callbacks = [
            *super().get_callbacks(),
            ModelCheckpoint(
                dirpath=self.checkpoint_folder,
                monitor=f"{ModelKey.VAL}/{self.primary_val_metric}",
                filename=self.best_checkpoint_filename,
                auto_insert_metric_name=False,
                mode="max" if self.maximise_primary_metric else "min",
            ),
        ]
        if self.analyse_loss:
            callbacks.append(
                LossAnalysisCallback(
                    outputs_folder=self.outputs_folder,
                    max_epochs=self.max_epochs,
                    patience=self.loss_analysis_patience,
                    epochs_interval=self.loss_analysis_epochs_interval,
                    num_slides_scatter=self.num_slides_scatter,
                    num_slides_heatmap=self.num_slides_heatmap,
                    save_tile_ids=self.save_tile_ids,
                    log_exceptions=self.log_exceptions,
                )
            )
        return callbacks

    def get_checkpoint_to_test(self) -> Path:
        """
        Returns the full path to a checkpoint file that was found to be best during training, whatever criterion
        was applied there. Note that the checkpoint file might not (yet) exist, for example in the case of
        preempted jobs.
        """
        checkpoint_path = Path(self.checkpoint_folder, self.best_checkpoint_filename_with_suffix)
        return checkpoint_path

    def get_dataloader_kwargs(self) -> dict:
        num_cpus = os.cpu_count()
        assert num_cpus is not None  # for mypy
        logging.info(f"os.cpu_count()={num_cpus}")
        num_devices = self.num_gpus_per_node()
        # We ensure num_devices is not 0 for non-GPU machines
        # to avoid division by zero error when computing `workers_per_gpu`
        workers_per_gpu = num_cpus // (num_devices or 1)
        workers_per_gpu = min(self.max_num_workers, workers_per_gpu)
        logging.info(f"Using {workers_per_gpu} data loader worker processes per GPU")
        dataloader_kwargs = dict(num_workers=workers_per_gpu, pin_memory=True)
        return dataloader_kwargs

    def get_transforms_dict(self, image_key: str) -> Optional[Dict[ModelKey, Union[Callable, None]]]:
        """Returns the image transforms that the training, validation, and test dataloaders should use.

        For reproducible results, those may need to be made deterministic via setting a fixed
        random see. See `SlidesDataModule` for an example how to achieve that."""
        return None

    def get_encoder_params(self) -> EncoderParams:
        return create_from_matching_params(self, EncoderParams)

    def get_data_module_for_runner_mode(self) -> HistoDataModule:
        """Gets the data module that the model is using from the private data_module field. If it is not set yet, it
        is set from either the training or the evaluation data module, depending on the runner mode.

        :return: A data module for either training or evaluation, depending on the runner mode.
        """
        if self.data_module is None:
            if self.runner_mode == RunnerMode.TRAIN:
                self.data_module = self.get_data_module()
            elif self.runner_mode == RunnerMode.EVAL_FULL:
                self.data_module = self.get_eval_data_module()
            else:
                raise ValueError(f"Unknown runner mode {self.runner_mode}")
        return self.data_module

    def get_num_classes(self) -> int:
        """Gets the number of classes that the model handles. In training mode, this is the number of classes in the
        training dataset. In evaluation mode, this is the number of classes in the test dataset.

        This method has a side effect: It creates the data module if it is not yet set in the `data_module` attribute.
        """
        if self.runner_mode == RunnerMode.EVAL_FULL:
            return self.get_data_module_for_runner_mode().test_dataset.n_classes
        else:
            return self.get_data_module_for_runner_mode().train_dataset.n_classes

    def get_class_weights(self) -> torch.Tensor:
        """Gets the class weights that the model should use. In training mode, this is the class weights from the
        training data module. In evaluation mode, this is a tensor with all ones (the class weights will be loaded
        from the checkpoint, so their value does not matter).

        :return: A tensor if the model is used for training, None otherwise.
        """
        if self.runner_mode == RunnerMode.EVAL_FULL:
            num_classes = self.get_num_classes()
            return torch.ones(2 if num_classes == 1 else num_classes)
        else:
            return self.get_data_module_for_runner_mode().class_weights

    def create_model(self) -> DeepMILModule:
        num_classes = self.get_num_classes()
        outputs_handler = self.get_outputs_handler()
        deepmil_module = DeepMILModule(
            label_column=self.get_label_column(),
            n_classes=num_classes,
            class_names=self.class_names,
            class_weights=self.get_class_weights(),
            outputs_folder=self.outputs_folder,
            encoder_params=self.get_encoder_params(),
            pooling_params=create_from_matching_params(self, PoolingParams),
            classifier_params=create_from_matching_params(self, ClassifierParams),
            optimizer_params=create_from_matching_params(self, OptimizerParams),
            outputs_handler=outputs_handler,
            analyse_loss=self.analyse_loss,
        )
        deepmil_module.transfer_weights(self.trained_weights_path)
        outputs_handler.set_slides_dataset_for_plots_handlers(self.get_slides_dataset())
        outputs_handler.set_extra_slides_dataset_for_plots_handlers(self.get_extra_slides_dataset_for_plotting())
        return deepmil_module

    def get_data_module(self) -> HistoDataModule:
        raise NotImplementedError

    def get_eval_data_module(self) -> HistoDataModule:
        raise NotImplementedError

    def get_slides_dataset(self) -> Optional[SlidesDataset]:
        return None

    def get_extra_slides_dataset_for_plotting(self) -> Optional[SlidesDataset]:
        return None

    def ignore_pl_warnings(self) -> None:
        # Pytorch Lightning prints a warning if the batch size is not consistent across all batches. The way PL infers
        # the batch size is not compatible with our data loaders. It searches for the first item in the batch that is a
        # tensor and uses its size[0] as the batch size. However, in our case, the batch is a list of tensors, so it
        # thinks that the batch size is the bag_size which can be different for each WSI in the batch. This is why we
        # ignore this warning to avoid noisy logs.
        warnings.filterwarnings("ignore", ".*Trying to infer the `batch_size` from an ambiguous collection.*")

    def get_label_column(self) -> str:
        """Slides and Tiles pipeline use different label columns."""
        raise NotImplementedError


class BaseMILTiles(BaseMIL):
    """BaseMILTiles is an abstract subclass of BaseMIL for running MIL experiments on tiles datasets. It is responsible
    for instantiating the full DeepMIL model in tiles settings. Subclasses should define their datamodules and
    configure experiment-specific parameters.
    """

    # Tiles Data module parameters:
    cache_mode: CacheMode = param.ClassSelector(
        default=CacheMode.MEMORY,
        class_=CacheMode,
        doc="The type of caching to perform: 'memory' (default), 'disk', or 'none'.",
    )
    precache_location: CacheLocation = param.ClassSelector(
        default=CacheLocation.CPU,
        class_=CacheLocation,
        doc="Whether to pre-cache the entire transformed dataset "
        "upfront and save it to disk and if re-load in cpu or "
        "gpu. Options: `none`,`cpu` (default), `gpu`",
    )

    def get_label_column(self) -> str:
        return self.get_data_module_for_runner_mode().train_dataset.label_column

    def setup(self) -> None:
        super().setup()
        # Fine-tuning requires tiles to be loaded on-the-fly, hence, caching is disabled by default.
        # When tune_encoder and is_caching are both set, below lines should disable caching automatically.
        if self.tune_encoder:
            self.is_caching = False
        if not self.is_caching:
            self.cache_mode = CacheMode.NONE
            self.precache_location = CacheLocation.NONE

    def get_dataloader_kwargs(self) -> dict:
        if self.is_caching:
            dataloader_kwargs = dict(num_workers=0, pin_memory=False)
        else:
            dataloader_kwargs = super().get_dataloader_kwargs()
        return dataloader_kwargs

    def get_transforms_dict(self, image_key: str) -> Dict[ModelKey, Union[Callable, None]]:
        if self.is_caching:
            encoder = create_from_matching_params(self, EncoderParams).get_encoder(self.outputs_folder)
            transform = Compose(
                [
                    LoadTilesBatchd(image_key, progress=False),
                    EncodeTilesBatchd(image_key, encoder, chunk_size=self.encoding_chunk_size),  # type: ignore
                ]
            )
        else:
            transform = LoadTilesBatchd(image_key, progress=False)  # type: ignore
        # in case the transformations for training contain augmentations, val and test transform will be different
        return {ModelKey.TRAIN: transform, ModelKey.VAL: transform, ModelKey.TEST: transform}


class BaseMILSlides(BaseMIL, TilingParams):
    """BaseSlidesMIL is an abstract subclass of BaseMIL for running MIL experiments on slides datasets. It is
    responsible for instantiating the full DeepMIL model in slides settings. Subclasses should define their datamodules
    and configure experiment-specific parameters.
    """

    def get_label_column(self) -> str:
        return SlideKey.LABEL



#  -------------------------------------------------------------------------------------------
#  Copyright (c) Microsoft Corporation. All rights reserved.
#  Licensed under the MIT License (MIT). See LICENSE in the repo root for license information.
#  -------------------------------------------------------------------------------------------
import logging
import param
from copy import deepcopy
from health_azure.logging import logging_section
from health_cpath.datasets.base_dataset import SlidesDataset
from health_cpath.preprocessing.tiff_conversion import AMPERSAND, TIFF_EXTENSION, UNDERSCORE, ConvertWSIToTiffd
from health_cpath.utils.naming import SlideKey
from monai.data.dataset import Dataset
from pathlib import Path
from torch.utils.data import DataLoader
from tqdm import tqdm
from typing import List, Optional
from tifffile.tifffile import COMPRESSION


class TiffConversionConfig(param.Parameterized):
    image_key: SlideKey = param.ClassSelector(
        class_=SlideKey,
        default=SlideKey.IMAGE,
        doc="The key of the image in the dataset. This is used to get the path of the src file.",
    )
    target_magnifications: Optional[List[float]] = param.List(
        default=[5.0],
        class_=float,
        doc="The magnifications that will be saved in the tiff files. Use None for all available magnifications.",
    )
    add_lowest_magnification: bool = param.Boolean(
        default=False,
        doc="If True, the lowest magnification will be saved in the tiff files in addition to the target "
        "magnifications. If False, only the specified magnifications will be saved. This is especially useful for "
        "costly computations that can be applied at a lower magnification.",
    )
    default_base_objective_power: Optional[float] = param.Number(
        default=None,
        doc="The objective power of the base magnification of the originale whole slide image. This is used to "
        "calculate the levels corresponding to the target magnifications. If None, the objective power will be "
        "extracted from the properties of the src file.",
    )
    replace_ampersand_by: str = param.String(
        default=UNDERSCORE,
        doc="The character that will replace the ampersand in the file name. & (as in H&E) can be problematic in some "
        "file systems. It is recommended to use _ instead.",
    )
    compression: COMPRESSION = param.ClassSelector(
        default=COMPRESSION.ADOBE_DEFLATE,
        class_=COMPRESSION,
        doc="The compression that will be used for the tiff files. Default is ADOBE_DEFLATE for lossless compression.",
    )
    tile_size: int = param.Integer(
        default=512, doc="The size of the tiles used to write the tiff file, defaults to 512."
    )
    num_workers: int = param.Integer(
        default=1,
        doc="The number of workers that will be used to convert the src files to tiff files. Defaults to 1.",
    )
    converted_dataset_csv: str = param.String(
        default="",
        doc="The name of the new dataset csv file that will be created for the converted data. If None, the default "
        "name of the original dataset will be used.",
    )
    min_file_size: int = param.Integer(
        default=0,
        doc="The minimum size of the tiff file in bytes. If the tiff file is smaller than this size, it will get "
        "overwritten. Defaults to 0.",
    )
    verbose: bool = param.Boolean(
        default=False,
        doc="If True, the progress of the conversion will be logged including src and tiff file sizes. "
        "Defaults to False.",
    )

    def get_transform(self, output_folder: Path) -> ConvertWSIToTiffd:
        """Get the transform that will be used to convert the src files to tiff files."""
        return ConvertWSIToTiffd(
            output_folder=output_folder,
            image_key=self.image_key,
            target_magnifications=self.target_magnifications,
            add_lowest_magnification=self.add_lowest_magnification,
            default_base_objective_power=self.default_base_objective_power,
            replace_ampersand_by=self.replace_ampersand_by,
            compression=self.compression,
            tile_size=self.tile_size,
            min_file_size=self.min_file_size,
            verbose=self.verbose,
        )

    def create_dataset_csv_for_converted_data(self, output_folder: Path) -> None:
        """Create a new dataset csv file for the converted data.

        :param dataset_df: The original dataset csv file.
        :param output_folder: The folder where the new dataset csv file will be saved.
        """
        new_dataset_df = deepcopy(self.slides_dataset.dataset_df)
        new_dataset_df[self.slides_dataset.image_column] = (
            new_dataset_df[self.slides_dataset.image_column]
            .str.replace(AMPERSAND, self.replace_ampersand_by)
            .map(lambda x: str(Path(x).with_suffix(TIFF_EXTENSION)))
        )
        new_dataset_file = (
            self.converted_dataset_csv if self.converted_dataset_csv else self.slides_dataset.default_csv_filename
        )
        new_dataset_path = output_folder / new_dataset_file
        new_dataset_df.to_csv(new_dataset_path, sep="\t" if new_dataset_path.suffix == ".tsv" else ",")
        logging.info(f"Saved new dataset tsv file to {new_dataset_path}")

    def __call__(self, dataloader: DataLoader) -> None:
        for _ in tqdm(dataloader, total=len(dataloader)):
            pass

    def run(self, dataset: SlidesDataset, output_folder: Path, wsi_subfolder: Optional[str] = None) -> None:
        """Run the conversion of the src files to tiff files.

        dataset: The slides dataset that contains the src wsi.
        output_folder: The folder where the tiff files will be saved.
        image_subfolder: The subfolder where the tiff files will be saved. If None, the tiff files will be saved in the
            root output folder.
        """
        self.slides_dataset: SlidesDataset = dataset
        if wsi_subfolder is not None:
            wsi_output_folder = output_folder / wsi_subfolder
            wsi_output_folder.mkdir(parents=True, exist_ok=True)
            logging.info(f"Whole slide images will be saved to subfolder {wsi_output_folder}")
        else:
            wsi_output_folder = output_folder
        transformed_dataset = Dataset(dataset, self.get_transform(wsi_output_folder))  # type: ignore
        dataloader = DataLoader(transformed_dataset, num_workers=self.num_workers, batch_size=1)
        with logging_section(f"Conversion of {len(dataset)} slides to tiff format to {output_folder}"):
            self(dataloader)
            self.create_dataset_csv_for_converted_data(output_folder)
